import numpy as np
import torch

from pymde import constraints
from pymde import preprocess
from pymde import recipes
from pymde import util
from pymde.functions import penalties
import pymde.testing as testing


def test_k_nearest_neighbors():
    data_matrix = np.array([[0.0], [1.0], [1.5], [1.75]])
    graph = preprocess.data_matrix.k_nearest_neighbors(data_matrix, k=2)
    edges = set(tuple(e) for e in graph.edges.cpu().numpy().tolist())
    # [2, 1], [3, 2], [3, 1], omitted because they are duplicates
    expected = set(
        tuple(e)
        for e in np.array([[0, 1], [0, 2], [1, 2], [1, 3], [2, 3]]).tolist()
    )
    assert edges == expected

    edges = graph.edges.cpu().numpy()
    # 2 if i and j are neighbors of each other,
    # 1 if i is neighbor of j but not vice versa
    testing.assert_allclose(np.array([1.0, 1.0, 2.0, 2.0, 2.0]), graph.weights)


@testing.cpu_and_cuda
def test_laplacian_embedding(device):
    torch.manual_seed(0)
    data_matrix = torch.randn(100, 10, device=device)

    util.seed(0)
    laplacian_emb = recipes.laplacian_embedding(
        data_matrix, device=device
    ).embed()

    util.seed(0)
    also_laplacian = recipes.preserve_neighbors(
        data_matrix,
        attractive_penalty=penalties.Quadratic,
        repulsive_penalty=None,
        device=device,
    ).embed()

    also_laplacian = util.align(source=also_laplacian, target=laplacian_emb)

    testing.assert_allclose(
        laplacian_emb.cpu().numpy(), also_laplacian.cpu().numpy(), atol=1e-3
    )


@testing.cpu_and_cuda
def test_anchor_initialization(device):
    n_items = 10

    util.seed(0)
    data_matrix = torch.randn(n_items, 5, device=device)

    anchors = torch.tensor([0, 1, 3], device=device)
    values = torch.tensor([2.0, 1.0, 3.0], device=device).reshape(3, 1)
    constraint = constraints.Anchored(anchors, values)

    # preserve neighbors
    mde = recipes.preserve_neighbors(
        data_matrix,
        embedding_dim=1,
        constraint=constraint,
        init="random",
        device=device,
    )
    testing.assert_allclose(
        mde._X_init[anchors].cpu().numpy(), values.cpu().numpy()
    )

    mde = recipes.preserve_neighbors(
        data_matrix,
        embedding_dim=1,
        constraint=constraint,
        init="quadratic",
        device=device,
    )
    testing.assert_allclose(
        mde._X_init[anchors].cpu().numpy(), values.cpu().numpy()
    )


@testing.cpu_and_cuda
def test_no_anchor_anchor_edges(device):
    util.seed(0)
    data_matrix = torch.randn(3, 2, device=device)

    anchors = torch.tensor([0, 1], device=device)
    values = torch.tensor([2.0, 3.0], device=device).reshape(2, 1)
    constraint = constraints.Anchored(anchors, values)

    mde = recipes.preserve_distances(
        data_matrix, embedding_dim=1, constraint=constraint, device=device
    )
    expected_edges = torch.tensor([[0, 2], [1, 2]], device=device)
    testing.assert_all_equal(
        expected_edges.cpu().numpy(), mde.edges.cpu().numpy()
    )

    mde = recipes.preserve_neighbors(
        data_matrix, embedding_dim=1, constraint=constraint, device=device
    )
    testing.assert_all_equal(
        expected_edges.cpu().numpy(), mde.edges.cpu().numpy()
    )


@testing.cpu_and_cuda
def test_neighbor_reproducibility(device):
    def _run_test(n_items):
        torch.manual_seed(0)
        Y = torch.rand((n_items, 128), device=device)

        prev_edges = None
        prev_weights = None
        for i in range(5):
            util.seed(0)
            mde = recipes.preserve_neighbors(Y, device=device)
            edges = mde.edges
            weights = mde.distortion_function.weights
            if prev_edges is not None:
                assert (edges == prev_edges).all()
                assert (weights == prev_weights).all()
            prev_edges = edges
            prev_weights = weights

    _run_test(36)
    _run_test(1001)


@testing.cpu_and_cuda
def test_distances_reproducibility(device):
    def _run_test(n_items):
        torch.manual_seed(0)
        Y = torch.rand((n_items, 128), device=device)

        prev_edges = None
        prev_deviations = None
        for i in range(5):
            util.seed(0)
            mde = recipes.preserve_distances(
                Y, max_distances=1e5, device=device
            )
            edges = mde.edges
            deviations = mde.distortion_function.deviations
            if prev_edges is not None:
                assert (edges == prev_edges).all()
                assert (deviations == prev_deviations).all()
            prev_edges = edges
            prev_deviations = deviations

    _run_test(36)
    _run_test(1001)



"""Losses: distortion functions derived from original deviations.

A vector distortion function :math:`f : \\mathbf{R}^{p} \\to \\mathbf{R}^p`
derived from original deviations has component functions

.. math::

    f_k(d_k) = \\ell(d_k, \\delta_k), \\quad k=1, \\ldots, p,

where
:math:`\\ell` is a loss function, :math:`\\delta_k` is a nonnegative deviation
or dissimilarity score, :math:`d_k` is an embedding distance,

When an MDE problem calls a distortion function, :math:`d_k` is the Euclidean
distance between the items paired by the k-th edge, so :math:`\\delta_k` should
be the original deviation associated with the k-th edge, and :math:`f_k(d_k)`
is the distortion associated with the edge.

The deviations can be interpreted as targets for the embedding distances:
the loss function is 0 when :math:`d_k = \\delta_k`, and positive otherwise.
So a deviation :math:`\\delta_k`` of 0 means that the items in the k-th edge
are the same, and the larger the deviation, the more dissimilar the items are.

Distortion functions are created in a vectorized or elementwise fashion. The
constructor takes a sequence (torch.Tensor) of deviations (target distances),
returning a callable object. The object takes a sequence of distances of the
same length as the weights, and returns a sequence of distortions, one for each
distance.

Some examples of losses inlcude:

.. code:: python3

    pymde.losses.Absolute
    pymde.losses.Quadratic
    pymde.losses.SoftFractional

**Example.**

.. code:: python3

    deviations = torch.tensor([1., 2., 3.])
    f = pymde.losses.Quadratic(deviations)

    distances = torch.tensor([2., 5., 4.])
    distortions = f(distances)
    # the distortions are (2 - 1)**2 == 1, (5 - 2)**2 == 9, (4 - 3)**2 == 1
    print(distortions)

prints

.. code:: python3

    torch.tensor([1., 9., 1.])
"""

from pymde import util
from pymde.functions.function import Function
import torch


class Quadratic(Function):
    """:math:`\\ell(d, \\delta) = (d - \\delta)^2`"""

    def __init__(self, deviations):
        super(Quadratic, self).__init__()
        self.deviations = util.to_tensor(deviations)

    def forward(self, distances):
        return (self.deviations - distances).pow(2)


class WeightedQuadratic(Function):
    """:math:`\\ell(d, \\delta) = \\frac{1}{\\delta^2} (d - \\delta)^2`

    If ``weights`` is not None, the coefficient then the
    coefficient :math:`1/\\delta^2` is replaced by the weights.
    """

    def __init__(self, deviations, weights=None):
        super(WeightedQuadratic, self).__init__()
        self.deviations = util.to_tensor(deviations)
        if weights is None:
            weights = 1.0 / deviations.pow(2)
        self.weights = util.to_tensor(weights, device=self.deviations.device)

    def forward(self, distances):
        return self.weights * (self.deviations - distances).pow(2)


class _ClippedQuadratic(Function):
    def __init__(self, deviations, threshold):
        super(_ClippedQuadratic, self).__init__()
        self.deviations = util.to_tensor(deviations)
        self.threshold = threshold

    def forward(self, distances):
        diff = self.deviations - distances
        return torch.min(diff**2, (self.threshold + 1) ** 2)


class Huber(Function):
    """
    .. math::

        \\ell(d, \\delta) = \\begin{cases}
            \\cdot (d - \\delta)^2 & d < \\text{threshold} \\\\
            \\text{threshold}(2(d - \\delta) - \\cdot \\text{threshold})
            & d \\geq \\text{threshold}
        \\end{cases}
    """

    def __init__(self, deviations, threshold):
        super(Huber, self).__init__()
        self.deviations = util.to_tensor(deviations)
        self.threshold = threshold

    def forward(self, distances):
        output = torch.zeros(
            distances.shape, dtype=distances.dtype, device=distances.device
        )
        diff = (self.deviations - distances).abs()
        lt_idx = diff < self.threshold
        output[lt_idx] = diff[lt_idx].pow(2)
        output[~lt_idx] = self.threshold * (2 * diff[~lt_idx] - self.threshold)
        return output


class Cubic(Function):
    """:math:`\\ell(d, \\delta) = (d - \\delta)^3`"""

    def __init__(self, deviations):
        super(Cubic, self).__init__()
        self.deviations = util.to_tensor(deviations)

    def forward(self, distances):
        return (self.deviations - distances).abs().pow(3)


class Power(Function):
    """:math:`\\ell(d, \\delta) = (d - \\delta)^{\\text{exponent}}`"""

    def __init__(self, deviations, exponent):
        super(Power, self).__init__()
        self.deviations = util.to_tensor(deviations)
        self.exponent = util.to_tensor(exponent, device=self.deviations.device)

    def forward(self, distances):
        return (self.deviations - distances).abs().pow(self.exponent)


class _WeightedPower(Function):
    def __init__(self, deviations, exponent, weights=None):
        super(_WeightedPower, self).__init__()
        self.deviations = util.to_tensor(deviations)
        if weights is None:
            weights = 1.0 / deviations.pow(2)
        self.exponent = util.to_tensor(exponent, device=self.deviations.device)
        self.weights = util.to_tensor(weights, device=self.deviations.device)

    def forward(self, distances):
        return self.weights * (self.deviations - distances).abs().pow(
            self.exponent
        )


class Absolute(Function):
    """:math:`\\ell(d, \\delta) = |d - \\delta|`"""

    def __init__(self, deviations):
        super(Absolute, self).__init__()
        self.deviations = util.to_tensor(deviations)

    def forward(self, distances):
        return (self.deviations - distances).abs()


class Logistic(Function):
    """:math:`\\ell(d, \\delta) = \\log(1 + \\exp(|d - \\delta|))`"""

    def __init__(self, deviations):
        super(Logistic, self).__init__()
        self.deviations = util.to_tensor(deviations)

    def forward(self, distances):
        diff = self.deviations - distances
        return torch.log(1.0 + torch.exp(diff.abs()))


class Fractional(Function):
    """:math:`\\ell(d, \\delta) = \\max(\\delta / d, d / \\delta)`"""

    def __init__(self, deviations):
        super(Fractional, self).__init__()
        self.deviations = util.to_tensor(deviations)

    def forward(self, distances):
        return (
            torch.max(self.deviations / distances, distances / self.deviations)
            - 1.0
        )


class SoftFractional(Function):
    """:math:`\\ell(d, \\delta) = \\frac{1}{\\gamma}\\log\\left( \\frac{\\exp(\\gamma \\delta/d) + \\exp(\\gamma d/\\delta)}{2\\exp(\\gamma)} \\right)`

    The parameter ``gamma`` controls how close this loss is to the
    fractional loss. The larger ``gamma`` is, the closer to the fractional
    loss.
    """  # noqa: E501

    def __init__(self, deviations, gamma=10.0):
        super(SoftFractional, self).__init__()
        self.deviations = util.to_tensor(deviations)
        self.gamma = util.to_tensor(gamma, device=self.deviations.device)
        if gamma <= 0.0:
            raise ValueError("gamma must be positive, received ", float(gamma))

    def forward(self, distances):
        stacked = torch.stack(
            (self.deviations / distances, distances / self.deviations)
        ).T
        return (
            1.0
            / self.gamma
            * (
                torch.logsumexp(self.gamma * stacked, dim=1)
                - (torch.log(torch.tensor(2.0)) + self.gamma)
            )
        )


class _Log1p(Function):
    def __init__(self, deviations, exponent):
        super(_Log1p, self).__init__()
        self.deviations = util.to_tensor(deviations)
        self.exponent = util.to_tensor(exponent, device=self.deviations.device)

    def forward(self, distances):
        return torch.log(1 + (distances - self.deviations).pow(self.exponent))



# /// script
# requires-python = ">=3.10"
# dependencies = [
#     "marimo",
#     "pymde",
#     "numpy",
#     "pandas < 3.0",
#     "scikit-learn",
# ]
# ///

import marimo

__generated_with = "0.19.10"
app = marimo.App()

with app.setup:
    import marimo as mo
    import numpy as np
    import pandas as pd
    import pymde
    from sklearn import preprocessing


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    This notebook preprocesses the raw ACS dataset to produce the preprocessed data embedded in the 'Counties' notebook.
    """)
    return


@app.cell
def _():
    dataset = pymde.datasets.counties()
    return (dataset,)


@app.cell
def _(dataset):
    X = dataset.county_dataframe
    y = dataset.voting_dataframe
    return X, y


@app.cell
def _(X):
    X
    return


@app.cell
def _():
    totalpop_col_idx = 3
    col_idxes_to_normalize = [4, 5, 31]
    col_idxes_to_log = [totalpop_col_idx, 12, 13, 14, 15, 16]
    col_idxes_to_div_by_100 = [idx for idx in range(6, 11 + 1)]
    col_idxes_to_div_by_100 = col_idxes_to_div_by_100 + [idx for idx in range(17, 30)]
    col_idxes_to_div_by_100 = col_idxes_to_div_by_100 + [idx for idx in range(32, 36 + 1)]
    return (
        col_idxes_to_div_by_100,
        col_idxes_to_log,
        col_idxes_to_normalize,
        totalpop_col_idx,
    )


@app.cell
def _(
    X,
    col_idxes_to_div_by_100,
    col_idxes_to_log,
    col_idxes_to_normalize,
    totalpop_col_idx,
):
    col_names = X.columns
    for col_idx in col_idxes_to_normalize + col_idxes_to_log:
        col_idx_name = col_names[col_idx]
        X[col_idx_name] = X[col_idx_name].astype(float)
    for _row_idx, _row_obj in X.iterrows():
        totalpop_idx_name = col_names[totalpop_col_idx]
        totalpop = float(_row_obj[totalpop_idx_name])
        for col_idx in col_idxes_to_normalize:
            col_idx_name = col_names[col_idx]
            col_val = float(_row_obj[col_idx_name])
            X.at[_row_idx, col_idx_name] = col_val / totalpop
        for col_idx in col_idxes_to_log:
            col_idx_name = col_names[col_idx]
            col_val = float(_row_obj[col_idx_name])
            X.at[_row_idx, col_idx_name] = np.log(col_val)
        for col_idx in col_idxes_to_div_by_100:
            col_idx_name = col_names[col_idx]
            col_val = float(_row_obj[col_idx_name])
            X.at[_row_idx, col_idx_name] = col_val / 100.0
    return


@app.cell
def _(X):
    for _row_idx, _row_obj in X.iterrows():
        X.at[_row_idx, 'MeanCommute'] = X.at[_row_idx, 'MeanCommute'] / 60.0
    return


@app.cell
def _(y):
    y.rename(columns={"combined_fips":"CountyId"}, inplace=True)
    return


@app.cell
def _(X, y):
    Xy = X.merge(y, on="CountyId")
    return (Xy,)


@app.cell
def _(Xy):
    drop_cols = ["Unnamed: 0", "state_abbr", "county_name", "per_point_diff", "diff", "total_votes"]
    Xy.drop(columns=drop_cols, inplace=True)
    return


@app.cell
def _(Xy):
    y_many_cols = ['votes_dem', 'votes_gop', 'per_dem', 'per_gop', 'CountyId', 'State', 'County']
    y_many = Xy[y_many_cols]
    y_many = y_many.copy()
    X_1 = Xy.drop(columns=y_many_cols)
    return X_1, y_many


@app.cell
def _(X_1):
    X_2 = preprocessing.scale(X_1.to_numpy())
    return (X_2,)


@app.cell
def _(y_many):
    y_1 = pd.DataFrame(y_many['votes_gop'] <= y_many['votes_dem'])
    y_1 = y_1.to_numpy(dtype=np.int32)
    y_1 = y_1.flatten()
    return


@app.cell
def _(y_many):
    y_prop = pd.DataFrame(y_many["per_dem"]).to_numpy(dtype=np.float64)[:,0]
    return (y_prop,)


@app.cell
def _(X_2, y_prop):
    np.save('democratic_fraction.npy', y_prop)
    np.save('county_data.npy', X_2)
    return


if __name__ == "__main__":
    app.run()



"""Distortion function base class

`Function` is just a torch.nn.Module that has a device property.
"""

from pymde import util
import torch


class Function(torch.nn.Module):
    def __init__(self):
        """Distortion function.

        Subclasses should implement the `forward` function, with signature

        forward(self, distances : torch.Tensor) -> distortion : torch.Tensor,

        which maps a vector of embedding distances to a vector of distortions,
        one for each distance.
        """
        super(Function, self).__init__()

    def __setattr__(self, name, value):
        if isinstance(value, torch.Tensor):
            self.register_buffer(name, value)
        else:
            super(Function, self).__setattr__(name, value)

    @property
    def device(self):
        return util._module_device(self)


class StochasticFunction(torch.nn.Module):
    def __init__(self, function, sampler, batch_size, device="cpu", p=None):
        super(StochasticFunction, self).__init__()
        self.function_factory = function
        self.sampler = sampler
        self.batch_size = batch_size
        self.device = util.canonical_device(device)
        self.p = p

    def sample(self):
        edges, args = self.sampler(self.batch_size)
        return self.function_factory(*args), edges



"""Tests for Rust NN-Descent implementation."""

import numpy as np

from pymde._native import nn_descent, knn_l2


def _brute_force_knn(data, k):
    """Ground truth using exact Rust knn_l2."""
    neighbors, sq_distances = knn_l2(data, k)
    return neighbors, np.sqrt(np.maximum(sq_distances, 0.0))


def _recall(approx_neighbors, gt_neighbors, k):
    """Fraction of true neighbors found, averaged over all queries."""
    n = gt_neighbors.shape[0]
    # Both arrays have k+1 columns (column 0 = self); compare columns 1:
    hits = 0
    for i in range(n):
        approx_set = set(approx_neighbors[i, 1 : k + 1])
        gt_set = set(gt_neighbors[i, 1 : k + 1])
        hits += len(approx_set & gt_set)
    return hits / (n * k)


class TestNNDescentSmall:
    """Small dataset tests (exact or near-exact results)."""

    def test_small_recall(self):
        rng = np.random.default_rng(42)
        data = rng.standard_normal((50, 8)).astype(np.float32)
        k = 5
        neighbors, distances = nn_descent(data, k, seed=42)
        gt_neighbors, _ = _brute_force_knn(data, k)
        r = _recall(neighbors, gt_neighbors, k)
        assert r >= 0.95, f"recall {r:.4f} < 0.95 for n=50"

    def test_output_shape(self):
        rng = np.random.default_rng(42)
        data = rng.standard_normal((50, 4)).astype(np.float32)
        k = 5
        neighbors, distances = nn_descent(data, k, seed=42)
        assert neighbors.shape == (50, k + 1)
        assert distances.shape == (50, k + 1)

    def test_output_dtypes(self):
        rng = np.random.default_rng(42)
        data = rng.standard_normal((50, 4)).astype(np.float32)
        neighbors, distances = nn_descent(data, 5, seed=42)
        assert neighbors.dtype == np.int64
        assert distances.dtype == np.float32

    def test_self_neighbor(self):
        rng = np.random.default_rng(42)
        data = rng.standard_normal((50, 4)).astype(np.float32)
        neighbors, distances = nn_descent(data, 5, seed=42)
        # Column 0 should be self
        np.testing.assert_array_equal(neighbors[:, 0], np.arange(50))
        np.testing.assert_array_equal(distances[:, 0], 0.0)

    def test_distances_sorted(self):
        rng = np.random.default_rng(42)
        data = rng.standard_normal((50, 8)).astype(np.float32)
        neighbors, distances = nn_descent(data, 10, seed=42)
        for i in range(50):
            assert all(
                distances[i, j] <= distances[i, j + 1]
                for j in range(distances.shape[1] - 1)
            ), f"row {i}: distances not sorted"

    def test_reproducibility(self):
        rng = np.random.default_rng(42)
        data = rng.standard_normal((100, 8)).astype(np.float32)
        nb1, d1 = nn_descent(data, 10, seed=123)
        nb2, d2 = nn_descent(data, 10, seed=123)
        np.testing.assert_array_equal(nb1, nb2)
        np.testing.assert_array_equal(d1, d2)


class TestNNDescentMedium:
    """Medium dataset tests."""

    def test_medium_recall(self):
        rng = np.random.default_rng(42)
        data = rng.standard_normal((5000, 32)).astype(np.float32)
        k = 15
        neighbors, distances = nn_descent(data, k, seed=42)
        gt_neighbors, _ = _brute_force_knn(data, k)
        r = _recall(neighbors, gt_neighbors, k)
        assert r >= 0.95, f"recall {r:.4f} < 0.95 for n=5000"

    def test_distances_match_recomputed(self):
        rng = np.random.default_rng(42)
        data = rng.standard_normal((200, 16)).astype(np.float32)
        neighbors, distances = nn_descent(data, 10, seed=42)
        # Recompute L2 distances for the reported neighbors
        for i in range(200):
            for j in range(1, 11):
                nb = neighbors[i, j]
                expected = np.sqrt(np.sum((data[i] - data[nb]) ** 2))
                np.testing.assert_allclose(
                    distances[i, j],
                    expected,
                    rtol=1e-4,
                    err_msg=f"row {i} col {j}: distance mismatch",
                )


class TestNNDescentLarge:
    """Larger dataset to exercise the auto-switching threshold."""

    def test_large_recall(self):
        rng = np.random.default_rng(42)
        data = rng.standard_normal((11000, 32)).astype(np.float32)
        k = 15
        neighbors, distances = nn_descent(data, k, seed=42, verbose=True)
        gt_neighbors, _ = _brute_force_knn(data, k)
        r = _recall(neighbors, gt_neighbors, k)
        assert r >= 0.95, f"recall {r:.4f} < 0.95 for n=11000"


class TestKNearestNeighborsIntegration:
    """Integration test: nn_descent can be called from Python directly."""

    def test_knn_integration(self):
        from pymde.preprocess.data_matrix import k_nearest_neighbors

        rng = np.random.default_rng(42)
        data = rng.standard_normal((5000, 32)).astype(np.float32)
        graph = k_nearest_neighbors(data, k=15, verbose=True)
        # Should return a valid Graph object
        assert graph.edges.shape[0] > 0
        assert graph.edges.shape[1] == 2



# /// script
# requires-python = ">=3.10"
# dependencies = [
#     "marimo",
#     "pymde",
#     "matplotlib",
# ]
# ///

import marimo

__generated_with = "0.19.10"
app = marimo.App()

with app.setup:
    import matplotlib.pyplot as plt
    import marimo as mo
    import numpy as np
    import pymde
    import torch


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    # Dissimilar edges and negative weights

    In this notebook, we show how the number of repulsive edges and the magnitude of the negative weights
    can affect the embeddings, when creating embeddings based on weights. We use MNIST as a case study.

    The summary of what we'll find: the more dissimilar edges (and the larger the magnitude of the negative weight), the more uniformly spread out the embedding is. The fewer edges (and the smaller the magnitude of the weight), the more tightly clustered the embedding is. These comments are for when a standardization constraint is enforced. Without the constraint, the embedding can collapse to 0 if there are too few dissimilar edges.

    You should make sure to look at the MNIST notebook before going through this one.
    """)
    return


@app.cell
def _():
    mnist = pymde.datasets.MNIST()

    images = mnist.data
    digits = mnist.attributes['digits']
    return digits, images


@app.cell
def _(images):
    knn_graph = pymde.preprocess.k_nearest_neighbors(images, k=15, verbose=True)
    return (knn_graph,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ### Quadratic MDE problem
    """)
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    We start by computing a spectral initialization.
    """)
    return


@app.cell
def _(images, knn_graph):
    quadratic_mde = pymde.MDE(
        n_items=images.shape[0],
        embedding_dim=2,
        edges=knn_graph.edges,
        distortion_function=pymde.penalties.Quadratic(knn_graph.weights),
        constraint=pymde.Standardized())
    quadratic_mde.embed(verbose=True)
    return (quadratic_mde,)


@app.cell
def _(digits, quadratic_mde):
    quadratic_mde.plot(color_by=digits)
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## Dissimilar edges
    """)
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    First we vary the number of dissimilar edges, and show what effect this has on the embedding.
    """)
    return


@app.cell
def _(images, knn_graph):
    def get_edges(n_dissimilar_edges, neg_weight=-1.0):
        similar_edges = knn_graph.edges
        dissimilar_edges = pymde.preprocess.dissimilar_edges(images.shape[0], num_edges=n_dissimilar_edges, similar_edges=similar_edges)
        _edges = torch.cat([similar_edges, dissimilar_edges])
        _weights = torch.cat([knn_graph.weights, neg_weight * torch.ones(dissimilar_edges.shape[0])])
        return (_edges, _weights)

    return (get_edges,)


@app.cell
def _(digits, get_edges, images, knn_graph, quadratic_mde):
    n_similar = knn_graph.edges.shape[0]
    n_dissimilar = [n_similar // 8, n_similar // 4, n_similar // 2, n_similar, 2 * n_similar, 4 * n_similar]
    _device = 'cuda' if torch.cuda.is_available() else 'cpu'
    for n_dissimilar_edges in n_dissimilar:
        _edges, _weights = get_edges(n_dissimilar_edges)
        print('# dissimilar edges / total: ', n_dissimilar_edges / _edges.shape[0])
        _f = pymde.penalties.PushAndPull(weights=_weights, attractive_penalty=pymde.penalties.Log1p, repulsive_penalty=pymde.penalties.Log).to(_device)
        _std_mde = pymde.MDE(n_items=images.shape[0], embedding_dim=2, edges=_edges.to(_device), distortion_function=_f, constraint=pymde.Standardized())
        _std_mde.embed(X=quadratic_mde.X.to(_device))
        _std_mde.plot(color_by=digits)
        plt.show()
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## Negative weights

    Next we vary the weights instead of the number of edges. The effect is similar to what happened when we varied the dissimilar edges.
    """)
    return


@app.cell
def _(digits, get_edges, images, knn_graph, quadratic_mde):
    neg_weights = [-1 / 8, -1 / 4, -1 / 2, -1.0, -2.0, -4.0]
    _device = 'cuda' if torch.cuda.is_available() else 'cpu'
    for neg_weight in neg_weights:
        print('negative weight: ', neg_weight)
        _edges, _weights = get_edges(knn_graph.edges.shape[0], neg_weight)
        _f = pymde.penalties.PushAndPull(weights=_weights, attractive_penalty=pymde.penalties.Log1p, repulsive_penalty=pymde.penalties.Log).to(_device)
        _std_mde = pymde.MDE(images.shape[0], embedding_dim=2, edges=_edges.to(_device), distortion_function=_f, constraint=pymde.Standardized())
        _std_mde.embed(X=quadratic_mde.X.to(_device))
        _std_mde.plot(color_by=digits)
        plt.show()
    return


if __name__ == "__main__":
    app.run()



# /// script
# requires-python = ">=3.10"
# dependencies = [
#     "ipython==9.10.0",
#     "marimo",
#     "numpy==2.3.5",
#     "pymde==0.2.3",
#     "scipy==1.17.0",
#     "torch==2.10.0",
# ]
# ///

import marimo

__generated_with = "0.19.11"
app = marimo.App()

with app.setup:
    import marimo as mo
    import numpy as np
    import pymde
    import scipy.sparse as sp
    import torch


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    # Drawing Graphs


    This notebook shows how to use `PyMDE` to layout graphs, using simple MDE problems.
    """)
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## Creating a graph

    The `PyMDE.Graph` class encapsulates an undirected weighted graph.

    Graphs have three parts: nodes, edges, and weights.

    - *Nodes* are represented by non-negative integers; a graph on $n$ nodes will have nodes $0, 1, 2, \ldots, n - 1$.

    - An *edge* is an array `e` of two node indices, with `e[0] < e[1]`.

    - Each edge has an associated scalar number, which can be interpreted as a *weight* on the edge or a distance between the nodes it connects.

    Graphs can be constructed from adjacency matrices or adjacency lists.
    - An *adjacency matrix* is a sparse matrix `A` of shape `(n, n)`, with `A[i, j]` giving the weight associated with edge $(i, j)$. If `A[i, j]` is `0`, then edge $(i, j)$ is not present in the grpah.
    - An *adjacency list* is an array of edges, with each row corresponding to an edge.

    Here is a simple example that show how to create graphs.
    """)
    return


@app.cell
def _():
    adjacency_matrix = sp.csr_matrix(np.array([[0, 1, 1], [1, 0, 1], [1, 1, 0]]))
    _triangle = pymde.Graph(adjacency_matrix)
    _triangle.draw()
    return


@app.cell
def _():
    _edges = torch.tensor([[0, 1], [0, 2], [1, 2]])
    _triangle = pymde.Graph.from_edges(_edges)
    _triangle.draw()
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## Drawing graphs

    The `draw` method lays out the graph in the Cartesian plane. Behind-the-scenes, it constructs and solves
    MDE problems that are suitable for laying out graphs.


    We used this method in the previous cells. In the remainder of this notebook, we show how to draw graphs by
    creating and solving MDE problems from scratch.
    """)
    return


@app.cell
def _():
    n_items = 20
    embedding_dim = 2
    return embedding_dim, n_items


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ### Complete Graph
    """)
    return


@app.cell
def _(embedding_dim, n_items):
    _edges = pymde.all_edges(n_items)
    mde = pymde.MDE(
        n_items,
        embedding_dim=embedding_dim,
        edges=_edges,
        distortion_function=pymde.penalties.Cubic(weights=1.0),
        constraint=pymde.Standardized(),
    )
    mde.embed()
    mde.plot(edges=_edges, background_color='white')
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ### Chain Graph
    """)
    return


@app.cell
def _(embedding_dim, n_items):
    chain_graph = pymde.Graph.from_edges(
        torch.tensor([(i, i + 1) for i in range(n_items - 1)])
    )
    _shortest_paths_graph = pymde.preprocess.graph.shortest_paths(chain_graph)
    mde_1 = pymde.MDE(
        n_items,
        embedding_dim,
        _shortest_paths_graph.edges,
        pymde.losses.WeightedQuadratic(deviations=_shortest_paths_graph.distances),
    )
    mde_1.embed()
    mde_1.plot(edges=chain_graph.edges, background_color="white")
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ### Cycle Graph
    """)
    return


@app.cell
def _(embedding_dim, n_items):
    cycle_graph = pymde.Graph.from_edges(
        torch.tensor([(i, i + 1) for i in range(n_items - 1)] + [(n_items - 1, 0)])
    )
    _shortest_paths_graph = pymde.preprocess.graph.shortest_paths(cycle_graph)
    mde_2 = pymde.MDE(
        n_items,
        embedding_dim,
        _shortest_paths_graph.edges,
        pymde.losses.WeightedQuadratic(_shortest_paths_graph.distances),
    )
    mde_2.embed()
    mde_2.plot(edges=cycle_graph.edges, background_color="white")
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ### Star Graph
    """)
    return


@app.cell
def _(embedding_dim, n_items):
    star_graph = pymde.Graph.from_edges(
        torch.tensor([(0, i) for i in range(1, n_items)])
    )
    _shortest_paths_graph = pymde.preprocess.graph.shortest_paths(star_graph)
    mde_3 = pymde.MDE(
        n_items,
        embedding_dim,
        _shortest_paths_graph.edges,
        pymde.losses.WeightedQuadratic(_shortest_paths_graph.distances),
    )
    mde_3.embed()
    mde_3.plot(edges=star_graph.edges, background_color="white")
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ### Tree
    """)
    return


@app.cell
def _(embedding_dim, n_items):
    _edges = []
    stack = [0]
    while stack:
        root = stack.pop()
        c1 = root * 2 + 1
        c2 = root * 2 + 2
        if c1 < n_items:
            _edges.append((root, c1))
            stack.append(c1)
        if c2 < n_items:
            _edges.append((root, c2))
            stack.append(c2)
    tree = pymde.Graph.from_edges(torch.tensor(_edges))
    _shortest_paths_graph = pymde.preprocess.graph.shortest_paths(tree)
    mde_4 = pymde.MDE(
        n_items,
        embedding_dim,
        _shortest_paths_graph.edges,
        pymde.losses.WeightedQuadratic(_shortest_paths_graph.distances),
    )
    mde_4.embed(snapshot_every=1, max_iter=20)
    mde_4.plot(edges=tree.edges, background_color='white')
    return mde_4, tree


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    The `play` method creates a movie showing how the embedding was made. Try it out!
    """)
    return


@app.cell
def _(mde_4, tree):
    mde_4.play(edges=tree.edges, fps=15.0, savepath="/tmp/graph.gif")
    mo.image("/tmp/graph.gif")
    return


if __name__ == "__main__":
    app.run()



"""Standardized quadratic embeddings from weights

Historical embeddings that reduce to eigenproblems, like PCA and spectral
embedding.
"""

import numpy as np
import scipy.sparse
import scipy.sparse.linalg
import torch

from pymde import problem
from pymde import util
from pymde.functions import penalties


@util.tensor_arguments
def pca(Y, embedding_dim):
    """PCA embedding of a data matrix.

    Arguments
    ---------
    Y: torch.Tensor(shape=(n, k))
        data matrix, with n >= k
    embedding_dim: int
        the number of eigenvectors to retrieve, the embedding dimension;
        must be <= k

    Returns
    -------
    torch.Tensor(shape=(n, embedding_dim))
        The top embedding_dim eigenvectors of YY^T, scaled by sqrt(n)
    """
    n = Y.shape[0]
    embedding_dim = embedding_dim.int()
    min_dim = min([n, Y.shape[1]])
    if embedding_dim > min_dim:
        raise ValueError(
            "Embedding dimension must be at most minimum dimension of Y"
        )

    # PCA requires the data to be centered.
    Y = Y - Y.mean(axis=0)[None, :]
    U, _, _ = torch.svd(Y)
    return np.sqrt(float(n)) * U[:, :embedding_dim]


def _laplacian(n, m, edges, weights, use_scipy=True):
    A = util.adjacency_matrix(n, m, edges, weights, use_scipy=use_scipy)
    L = -A
    if use_scipy:
        L.setdiag(np.asarray(A.sum(axis=1)).squeeze())
    else:
        # diag is currently 0
        diag_vals = torch.sparse.sum(A, dim=1).to_dense()
        diag_inds = torch.tensor(
            [[i, i] for i in range(len(A))], device=L.device, dtype=L.dtype
        )
        diag_inds = diag_inds.transpose(0, 1)
        diag_matrix = torch.sparse_coo_tensor(
            diag_inds,
            diag_vals,
            size=(n, n),
            dtype=torch.float32,
            device=L.device,
        )
        L = L + diag_matrix
        L = L.coalesce()
    return L


def _spectral(
    L,
    m,
    cg=False,
    max_iter=400,
    edges=None,
    weights=None,
    warm_start=False,
    device=None,
):
    n = L.shape[0]
    k = m + 1
    if not cg:
        num_lanczos_vectors = max(2 * k + 1, int(np.sqrt(L.shape[0])))
        eigenvalues, eigenvectors = scipy.sparse.linalg.eigsh(
            L,
            k,
            which="SM",
            ncv=num_lanczos_vectors,
            tol=1e-4,
            maxiter=L.shape[0] * 5,
        )
        order = np.argsort(eigenvalues)[1:k]
    else:
        if warm_start:
            mde = problem.MDE(
                n,
                k,
                edges,
                distortion_function=penalties.Quadratic(weights),
                device=device,
            )
            X_init = mde.embed(max_iter=40)
        else:
            X_init = util.proj_standardized(
                torch.randn((n, k), device=device),
                demean=True,
            )
        eigenvalues, eigenvectors = torch.lobpcg(
            A=L,
            X=X_init,
            tol=None,
            # largest: find the smallest eigenvalues
            largest=False,
            niter=max_iter,
        )
        order = torch.argsort(eigenvalues)[1:k]
    return eigenvectors[:, order]


# TODO(akshayka): Deprecate device argument, infer from edges/weights
def spectral(
    n_items, embedding_dim, edges, weights, cg=False, max_iter=40, device="cpu"
):
    r"""Compute a spectral embedding


    Solves the quadratic MDE problem

    .. math::

        \begin{array}{ll}
        \mbox{minimize} & \sum_{(i, j) in \text{edges}} w_{ij} d_{ij}^2 \\
        \mbox{subject to} & (1/n) X^T X = I, \quad d_{ij} = |x_i - x_j|_2.
        \end{array}

    By default, the problem is solved using a Lanczos method. If cg=True,
    LOBPCG is used; LOBPCG is warm-started by running a projected quasi-newton
    method for a small number of iterations. Use cg=True when the number
    of edges is very large, and when an approximate solution is satisfactory
    (the Lanczos method typically gives much more accurate solutions, but can
    be slower).

    Arguments
    ---------
    n_items: int
        The number of items
    embedding_dim: int
        The embedding dimension
    edges: torch.Tensor(shape=(n_edges, 2))
        A list of edges (i, j), 0 <= i < j < n_items
    weights: torch.Tensor(shape=(n_edges,))
        A list of nonnegative weights associated with each edge
    cg: bool
        If True, uses a preconditioned CG method to find the embedding,
        which requires that the Laplacian matrix plus the identity is
        positive definite; otherwise, a Lanczos method is used. Use True when
        the Lanczos method is too slow (which might happen when the number of
        edges is very large).
    max_iter: int
        max iteration count for the CG method
    device: str (optional)
        The device on which to allocate the embedding

    Returns
    -------
    torch.Tensor(shape=(n_items, embedding_dim))
        A spectral embedding, projected onto the standardization constraint
    """
    # use torch sparse and linalg for lobpcg
    use_scipy = not cg
    L = _laplacian(n_items, embedding_dim, edges, weights, use_scipy=use_scipy)
    emb = _spectral(L, embedding_dim, cg=cg, device=device, max_iter=max_iter)
    if use_scipy:
        emb -= emb.mean(axis=0)
        emb = torch.tensor(emb, dtype=weights.dtype, device=device)
    else:
        emb -= emb.mean(dim=0)
    return util.proj_standardized(emb)



import matplotlib
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable

import numpy as np
import torch


def latexify(figsize_inches=None, font_size=12):
    """Set up matplotlib's RC params for LaTeX plotting.

    This function only needs to be called once per Python session.

    Arguments
    ---------
    figsize_inches: tuple float (optional)
        width, height of figure on inches

    font_size: int
        Size of font.
    """

    usetex = matplotlib.checkdep_usetex(True)
    if not usetex:
        raise RuntimeError(
            "Matplotlib could not find a LaTeX installation on your machine."
        )

    # code adapted from http://www.scipy.org/Cookbook/Matplotlib/LaTeX_Examples
    fig_width, fig_height = (
        figsize_inches if figsize_inches is not None else (None, None)
    )
    if fig_width is None:
        fig_width = 3.39

    if fig_height is None:
        golden_mean = (np.sqrt(5) - 1.0) / 2.0  # aesthetic ratio
        fig_height = fig_width * golden_mean  # height in inches

    max_height_inches = 8.0
    if fig_height > max_height_inches:
        print(
            "warning: fig_height too large:"
            + fig_height
            + "so will reduce to"
            + max_height_inches
            + "inches."
        )
        fig_height = max_height_inches

    params = {
        "backend": "ps",
        "text.latex.preamble": "\\usepackage{gensymb}",
        "axes.labelsize": font_size,
        "axes.titlesize": font_size,
        "font.size": font_size,
        "legend.fontsize": font_size,
        "xtick.labelsize": font_size,
        "ytick.labelsize": font_size,
        "text.usetex": True,
        "figure.figsize": [fig_width, fig_height],
        "font.family": "serif",
    }

    matplotlib.rcParams.update(params)


def _format_axes(ax, spine_color="black", linewidth=0.7):

    for spine in ["left", "bottom", "top", "right"]:
        ax.spines[spine].set_color(spine_color)
        ax.spines[spine].set_linewidth(linewidth)

    # ax.xaxis.set_ticks_position('bottom')

    for axis in [ax.xaxis, ax.yaxis]:
        axis.set_tick_params(direction="out", color=spine_color)

    return ax


def _cdf(array, n_bins=None, figsize_inches=(8, 3), xlabel=None):
    array = np.sort(array)
    if n_bins is None:
        n_bins = min(array.size, 100000)
    step = int(array.size / n_bins)
    bins = np.arange(1, array.size + 1, step=step)
    if bins[-1] != array.size:
        bins = np.concatenate((bins, np.array([array.size])))
    heights = bins / float(array.size)

    plt.figure(figsize=figsize_inches)
    ax = plt.gca()
    ax.step(array[bins - 1], heights, color="tab:blue")

    nonzero_index = np.argmax(array > 0)
    if nonzero_index > 0 or array[0] > 0:
        # array is positive, can take log
        ax.set_xlim(left=array[nonzero_index], right=None)
        plt.xscale("log")

    if xlabel is not None:
        plt.xlabel(xlabel)
    _format_axes(ax)
    plt.tight_layout()


def _show_cdf(mde, n_bins=None, X=None, figsize_inches=(8, 3)):
    X = X if X is not None else mde.X
    distortions = mde.distortions(X).cpu().numpy()
    _cdf(
        distortions,
        n_bins=n_bins,
        figsize_inches=figsize_inches,
        xlabel="distortions",
    )


def _labels_to_codes(labels):
    unique_labels = sorted(list(set(labels)))
    label2code = {label: code for code, label in enumerate(unique_labels)}
    codes = [label2code[label] for label in labels]
    return codes, unique_labels


def _square_lim(lim_low, lim_high):
    if abs(lim_low) > abs(lim_high) and lim_low < 0:
        return (lim_low, -lim_low)
    elif abs(lim_low) <= abs(lim_high) and lim_high > 0:
        return (-lim_high, lim_high)
    else:
        return (lim_low, lim_high)


def _is_discrete(dtype):
    return any(
        [
            np.issubdtype(dtype, other)
            for other in (
                np.integer,
                np.bool_,
                np.bytes_,
                np.str_,
                np.object_,
            )
        ]
    )


def _plot_3d(
    X, color_by, cmap, colors, edges, s, background_color, figsize, lim
):

    if isinstance(X, torch.Tensor):
        X = X.cpu().numpy()

    shadowsize = 0.03
    shadowcolor = "gainsboro"

    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(projection="3d")

    x, y, z = X[:, 0], X[:, 1], X[:, 2]

    if color_by is not None:
        discrete = _is_discrete(color_by.dtype)
        if discrete:
            c, unique_labels = _labels_to_codes(color_by)
        else:
            c, unique_labels = color_by, None
    elif colors is not None:
        c = colors
        cmap = None
    else:
        c = None
        cmap = None
    im = ax.scatter(
        x,
        y,
        z,
        c=c,
        edgecolor="k",
        cmap=cmap,
        alpha=1.0,
        s=s,
        linewidth=s / 20.0,
    )

    if lim is None:
        lim_low = min(np.min(x), np.min(y), np.min(z)) * 1.1
        lim_high = max(np.max(x), np.max(y), np.max(z)) * 1.1
        lim = _square_lim(lim_low, lim_high)
    ax.set_xlim(lim)
    ax.set_ylim(lim)
    ax.set_zlim(lim)

    if edges is None:
        # shadows
        ax.scatter(
            y,
            z,
            zdir="x",
            zs=ax.axes.get_xlim3d()[0],
            c=shadowcolor,
            alpha=0.5,
            marker="o",
            s=shadowsize,
        )
        ax.scatter(
            x,
            y,
            zdir="z",
            zs=ax.axes.get_zlim3d()[0],
            c=shadowcolor,
            alpha=0.5,
            marker="o",
            s=shadowsize,
        )

    if background_color is not None:
        if isinstance(background_color, str):
            bg = matplotlib.colors.to_rgba(background_color)
        else:
            bg = background_color
    elif edges is not None:
        bg = matplotlib.colors.to_rgba("k")
    else:
        bg = None

    if bg is not None:
        ax.xaxis.set_pane_color(bg)
        ax.yaxis.set_pane_color(bg)
        ax.zaxis.set_pane_color(bg)

    if edges is not None:
        if isinstance(edges, torch.Tensor):
            edges = edges.cpu().numpy()
        linewidth = 4.0 / np.log(edges.shape[0])
        edge_color = "w" if background_color is None else "k"
        for e in edges:
            xi = X[e[0]]
            xj = X[e[1]]
            ax.plot(
                [xi[0], xj[0]],
                [xi[1], xj[1]],
                [xi[2], xj[2]],
                color=edge_color,
                alpha=0.5,
                linewidth=linewidth,
                zorder=5,
            )

    if color_by is not None:
        if discrete:
            cbar = fig.colorbar(
                im,
                boundaries=np.arange(len(set(color_by)) + 1) - 0.5,
                shrink=0.5,
                orientation="horizontal",
                anchor=(0.5, 1.8),
            )
            cbar.set_ticks(np.arange(len(unique_labels)))
            if isinstance(unique_labels[0], str):
                rotation = 90
            else:
                rotation = 0
            cbar.ax.set_xticklabels(unique_labels, rotation=rotation)
        else:
            cbar = fig.colorbar(
                im,
                shrink=0.5,
                orientation="horizontal",
                anchor=(0.5, 1.8),
            )

    return ax


def _plot(
    X=None,
    color_by=None,
    edges=None,
    s=1.0,
    figsize=(6.4, 6.0),
    colors=None,
    cmap="Spectral",
    lim=None,
    background_color=None,
    title=None,
    tight=True,
    savepath=None,
):
    if isinstance(X, torch.Tensor):
        X = X.cpu().numpy()
    else:
        X = np.asarray(X)

    if isinstance(color_by, torch.Tensor):
        color_by = color_by.cpu().numpy()
    elif color_by is not None:
        color_by = np.asarray(color_by)

    if isinstance(edges, torch.Tensor):
        edges = edges.cpu().numpy()
    elif edges is not None:
        edges = np.asarray(edges)

    if isinstance(colors, torch.Tensor):
        colors = colors.cpu().numpy()
    elif colors is not None:
        colors = np.asarray(colors)

    X = X[:, :3]
    if X.shape[1] == 3:
        ax = _plot_3d(
            X,
            lim=lim,
            color_by=color_by,
            cmap=cmap,
            colors=colors,
            edges=edges,
            s=s,
            background_color=background_color,
            figsize=figsize,
        )
    else:
        if X.shape[1] == 1:
            X = np.hstack([X, np.zeros((X.shape[0], 1))])

        fig, ax = plt.subplots(1, 1, figsize=figsize)
        if background_color is not None:
            ax.set_facecolor(background_color)

        if lim is not None:
            ax.set_xlim(lim)
            ax.set_ylim(lim)
        else:
            lim_low = min(np.min(X[:, 0]), np.min(X[:, 1])) * 1.1
            lim_high = max(np.max(X[:, 0]), np.max(X[:, 1])) * 1.1
            lim = _square_lim(lim_low, lim_high)
            ax.set_xlim(lim)
            ax.set_ylim(lim)
            ax.set_aspect("equal", adjustable="box")

        if edges is not None:
            linewidth = 6.0 / np.log(edges.shape[0])
            scatter_color = "w" if background_color is None else "k"
            for e in edges:
                ax.plot(
                    [X[e[0], 0], X[e[1], 0]],
                    [X[e[0], 1], X[e[1], 1]],
                    linestyle="-",
                    markersize=s / 2.0,
                    color=scatter_color,
                    alpha=0.5,
                    linewidth=linewidth,
                )
            if background_color is None:
                ax.set_facecolor("k")

        # scatterplot of embedding
        if color_by is not None:
            discrete = _is_discrete(color_by.dtype)
            if discrete:
                codes, unique_labels = _labels_to_codes(color_by)
            else:
                codes = color_by.data
                unique_labels = None

            # edgecolor=[] prevents small points from having no fill
            im = ax.scatter(
                X[:, 0],
                X[:, 1],
                s=s,
                c=codes,
                edgecolor=[],
                cmap=cmap,
                alpha=1.0,
            )
        elif colors is not None:
            im = ax.scatter(
                X[:, 0],
                X[:, 1],
                s=s,
                c=colors,
                edgecolor=[],
                alpha=1.0,
            )
        else:
            ax.scatter(X[:, 0], X[:, 1], s=s, alpha=1.0)

        if title is not None:
            ax.set_title(title)

        _format_axes(ax)
        if tight:
            plt.tight_layout()

        if color_by is not None:
            aspect = 30
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="3.%", pad=0.05)
            if discrete:
                n_labels = len(unique_labels)
                cbar = fig.colorbar(
                    im,
                    ax=ax,
                    cax=cax,
                    boundaries=np.arange(n_labels + 1) - 0.5,
                    aspect=aspect,
                )
                cbar.set_ticks(np.arange(n_labels))
                cbar.ax.set_yticklabels(unique_labels)
            else:
                cbar = fig.colorbar(
                    im,
                    ax=ax,
                    cax=cax,
                    aspect=aspect,
                )

    if savepath is not None:
        fig.savefig(savepath, bbox_inches="tight")

    return ax


def plot(
    X,
    color_by=None,
    color_map="Spectral",
    colors=None,
    edges=None,
    axis_limits=None,
    background_color=None,
    marker_size=1.0,
    figsize_inches=(8.0, 8.0),
    savepath=None,
):
    """Plot an embedding, in one, two, or three dimensions.

    This function plots embeddings. The input embedding's dimension should
    be at most 3.

    The embedding is visualized as a scatter plot. The points can
    optionally be colored according to categorical or continuous values,
    or according to a pre-defined sequence of colors. Additionally,
    edges can optionally be superimposed.

    Arguments
    ---------
    X: array-like
        The embedding to plot, of shape ``(n_items, embedding_dim)``. The
        second dimension should be 1, 2, or 3.
    color_by: array-like, optional
        A sequence of values, one for each item, which should be
        used to color each embedding vector. These values may either
        be categorical or continuous. For example, if ``n_items`` is 4,

        .. code:: python3

            np.ndarray(['dog', 'cat', 'zebra', 'cat'])
            np.ndarray([0, 1, 1, 2]
            np.ndarray([0.1, 0.5, 0.31, 0.99]

        are all acceptable. The first two are treated as categorical,
        the third is continuous. A finite number of colors is used
        when the values are categorical, while a spectrum of colors is
        used when the values are continuous.
    color_map: str or matplotlib colormap instance
        Color map to use when resolving ``color_by`` to colors; ignored
        when ``color_by`` is None.
    colors: array-like, optional
        A sequence of colors, one for each item, specifying the exact
        color each item should be colored. Each row must represent
        an RGBA value.

        Only one of ``color_by`` and ``colors`` should be non-None.
    edges: array-like, optional
        List of edges to superimpose over the scatter plot, shape ``(any, 2)``
    axis_limits: tuple, optional
        tuple ``(limit_low, limit_high)`` of axis limits, applied to both
        the x and y axis.
    background_color: str, optional
        color of background
    marker_size: float, optional
        size of each point in the scatter plot
    figsize_inches: tuple
        size of figures in inches: ``(width_inches, height_inches)``
    savepath: str, optional
        path to save the plot.

    Returns
    -------
    matplotlib.Axes:
        Axis on which the embedding is plotted.
    """
    if color_by is not None and colors is not None:
        raise ValueError("Only one of 'color_by` and `colors` can be non-None")

    ax = _plot(
        X=X,
        color_by=color_by,
        cmap=color_map,
        colors=colors,
        edges=edges,
        lim=axis_limits,
        background_color=background_color,
        s=marker_size,
        figsize=figsize_inches,
    )

    if savepath is not None:
        plt.savefig(savepath)
    return ax


def _plot_gif_frame(
    snapshot,
    color_by,
    cmap,
    colors,
    edges,
    lim,
    background_color,
    figsize_inches,
    outdir,
    title,
    i,
    s,
):
    _plot(
        X=snapshot,
        color_by=color_by,
        cmap=cmap,
        edges=edges,
        title=title,
        lim=lim,
        background_color=background_color,
        tight=False,
        s=s,
        figsize=figsize_inches,
    )
    plt.savefig(outdir + "/{:04d}.png".format(i), dpi=144)
    plt.close("all")



"""Miscellenous preprocessing utilities."""

import numpy as np
import scipy.sparse as sp
import torch


from pymde import util


def sample_edges(n, num_edges, exclude=None, seed=None):
    """Randomly sample num_edges edges (optionally) not in excluded edges

    Arguments
    ---------
    n : int
        The number of items
    num_edges : int
        The (maximum) number of edges to sample. Must be at most (n choose 2).
    exclude: torch.Tensor, shape (-1, 2) (optional)
        edges to exclude from the sampled edges
    seed: int (optional)
        Random seed.

    Returns
    -------
    torch.Tensor
        randomly sampled edges; the edges are de-duplicated and ordered
        so that e[0] < e[1] for sampled edge e, so the number of rows
        in this tensor may be less than num_edges
    """
    if isinstance(exclude, torch.Tensor):
        exclude = exclude.cpu().numpy()

    n_all_edges = n * (n - 1) / 2
    n_excluded = 0 if exclude is None else exclude.shape[0]
    if num_edges > n_all_edges - n_excluded:
        raise ValueError(
            f"Cannot sample more than ({n} choose 2) - {n_excluded} ="
            f"{int(n * (n - 1) / 2.0) - n_excluded} edges. "
            f"(requested: {num_edges} edges)"
        )

    if exclude is not None and (exclude.shape[0] / n_all_edges) >= 0.2:
        A = sp.coo_matrix(
            (np.ones(exclude.shape[0]), (exclude[:, 0], exclude[:, 1])),
            shape=(n, n),
        ).todense()
        A[np.tril_indices(A.shape[0])] = 1.0
        A_comp = sp.coo_matrix(1 - A)
        sampled_edges = np.stack([A_comp.row, A_comp.col], axis=1)
        idx = np.random.choice(sampled_edges.shape[0], num_edges, replace=False)
        return torch.tensor(sampled_edges[idx])

    # randomly sample edges (i, j) with i < j via a bijection to
    # triangular numbers
    if seed is not None:
        randomstate = np.random.default_rng(seed)
    else:
        randomstate = util.np_rng()
    edge_idx = randomstate.choice(
        int(n * (n - 1) / 2), num_edges, replace=False, shuffle=False
    )
    u = (
        n
        - 2
        - np.floor(np.sqrt(-8 * edge_idx + 4 * n * (n - 1) - 7) / 2.0 - 0.5)
    )
    v = edge_idx + u + 1 - n_all_edges + (n - u) * ((n - u) - 1) / 2
    sampled_edges = np.stack([u, v], axis=1).astype(np.int64)

    if exclude is not None:
        # remove duplicated edges
        all_edges = np.concatenate([exclude, sampled_edges])
        neg_edge_offset = exclude.shape[0]
        unique, idx = np.unique(all_edges, axis=0, return_index=True)
        idx = np.sort(idx)
        cutoff = np.searchsorted(idx, neg_edge_offset)
        sampled_edges = all_edges[idx[cutoff:]]
    return torch.tensor(sampled_edges)


def dissimilar_edges(n_items, similar_edges, num_edges=None, seed=None):
    """Sample edges not in ``similar_edges``.

    Given a number of items, and a ``torch.Tensor`` containing pairs of
    items known to be similar, this function samples (uniformly at random) some
    edges in the complement of ``similar_edges``. The returned number
    of edges will be approximately equal to (and no greater than) the
    number of edges in ``similar_edges`` (or ``num_edges``, if provided).

    Arguments
    ---------
    n_items: int
        The number of items.
    similar_edges: torch.Tensor
        Edges to exclude when sampling.
    num_edges: int (optional)
        Number of edges to sample, defaults to approximately
        the same number as ``similar_edges``.
    seed: int (optional)
        Random seed for the sampling.

    Returns
    -------
    torch.Tensor
        Edges not in ``similar_edges``.
    """
    if num_edges is None:
        num_edges = similar_edges.shape[0]
    return sample_edges(n_items, num_edges, exclude=similar_edges, seed=seed)


def deduplicate_edges(edges):
    if isinstance(edges, torch.Tensor):
        device = edges.device
        dtype = edges.dtype
        edges = edges.cpu().numpy()
    else:
        device = None
        dtype = None

    flip_idx = edges[:, 0] > edges[:, 1]
    if flip_idx.any():
        edges[flip_idx] = np.stack(
            [edges[flip_idx][:, 1], edges[flip_idx][:, 0]], axis=1
        )
    unique_edges = np.unique(edges, axis=0)
    return torch.tensor(unique_edges, dtype=dtype, device=device)


def _rms(distances):
    return distances.pow(2).mean().sqrt()


def scale(distances, natural_length):
    alpha = natural_length / _rms(distances)
    return alpha * distances



# /// script
# requires-python = ">=3.10"
# dependencies = [
#     "sphinx",
#     "furo",
#     "autoclasstoc",
#     "pymde",
# ]
#
# [tool.uv.sources]
# pymde = { path = "..", editable = true }
# ///
"""Build the pymde documentation using Sphinx."""

import shutil
import subprocess
import sys
from pathlib import Path

DOCS_SRC = Path(__file__).resolve().parent.parent / "docs_src"
SOURCE_DIR = DOCS_SRC / "source"
BUILD_DIR = DOCS_SRC / "build"
OUTPUT_DIR = DOCS_SRC.parent / "docs"
CNAME_SRC = DOCS_SRC / "_CNAME"


def main():
    try:
        import pymde  # noqa: F401
    except ImportError:
        print(
            "Error: pymde must be installed for autodoc to work.\n"
            "Install it in development mode: pip install -e '.[dev]'",
            file=sys.stderr,
        )
        sys.exit(1)

    result = subprocess.run(
        [
            sys.executable, "-m", "sphinx",
            "-M", "html",
            str(SOURCE_DIR),
            str(BUILD_DIR),
        ],
        cwd=DOCS_SRC,
    )
    if result.returncode != 0:
        sys.exit(result.returncode)

    html_dir = BUILD_DIR / "html"

    # Mark for GitHub Pages (no Jekyll processing)
    (html_dir / ".nojekyll").touch()

    # Replace docs/ with freshly built HTML
    if OUTPUT_DIR.exists():
        shutil.rmtree(OUTPUT_DIR)
    shutil.move(str(html_dir), str(OUTPUT_DIR))

    # Custom domain CNAME
    if CNAME_SRC.exists():
        shutil.copy2(str(CNAME_SRC), str(OUTPUT_DIR / "CNAME"))

    print(f"Documentation built in {OUTPUT_DIR}")


if __name__ == "__main__":
    main()



# /// script
# requires-python = ">=3.10"
# dependencies = [
#     "marimo",
#     "pymde",
#     "matplotlib",
#     "torch",
# ]
# ///

import marimo

__generated_with = "0.19.11"
app = marimo.App()

with app.setup:
    import matplotlib
    import marimo as mo
    import pymde
    import torch


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    # Updating embeddings

    This notebook shows how to add new points to an existing embedding, using MNIST as an example.

    The basic idea is to use an anchor constraint to pin the existing embedding in place, then embed the new points.
    """)
    return


@app.cell
def _():
    mnist = pymde.datasets.MNIST()
    return (mnist,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    We'll start by embedding half of the MNIST data, using the preserve neighbors function.
    """)
    return


@app.cell
def _(mnist):
    n_train = 35000
    mnist_train = mnist.data[:n_train]
    mnist_train_labels = mnist.attributes['digits'][:n_train]

    mde = pymde.preserve_neighbors(mnist_train, constraint=pymde.Standardized(), verbose=True)
    mde.embed(verbose=True)
    mde.plot(color_by=mnist_train_labels)
    return mde, n_train


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## Updating the embedding

    Next, we'll augment the above embedding with embedding vectors for the rest of the MNIST dataset.

    We start by making an **anchor constraint**. This constraint says that we should pin the first `n_train` items to the embedding vectors that we just computed.
    """)
    return


@app.cell
def _(mde, n_train):
    anchor_constraint = pymde.Anchored(
        anchors=torch.arange(n_train),
        values=mde.X,
    )
    return (anchor_constraint,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    We now make a new MDE problem, passing in the entire MNIST dataset to the `preserve_neighbors` function. Crucially,
    we also provide the anchor constraint.
    """)
    return


@app.cell
def _(anchor_constraint, mnist):
    incremental_mde = pymde.preserve_neighbors(
        mnist.data,
        constraint=anchor_constraint,
        init='random',
        verbose=True)
    return (incremental_mde,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    The last step is to call the embed method. This will optimize the placement of the embedding vectors of the new images, while keeping the old ones fixed in place.
    """)
    return


@app.cell
def _(incremental_mde):
    incremental_mde.embed(eps=1e-6, verbose=True)
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    The first `n_train` rows of the new embedding are identitical to the old embedding, as verified below.
    """)
    return


@app.cell
def _(incremental_mde, mde, n_train):
    (incremental_mde.X[:n_train] == mde.X).all()
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    # Plot the final embedding

    Below we plot the entire embedding. Notice that it looks very similar to the first embedding. This suggests that the incremental embedding was effective.
    """)
    return


@app.cell
def _(incremental_mde, mnist):
    pymde.plot(incremental_mde.X, color_by=mnist.attributes['digits'])
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    We can compare just the new embedding vectors with the old embedding vectors. Note that both plots are nearly identitcal.
    """)
    return


@app.cell
def _(incremental_mde, mnist, n_train):
    _ax = pymde.plot(incremental_mde.X[n_train:], color_by=mnist.attributes['digits'][n_train:])
    _ax.set_title('New embedding vectors')
    return


@app.cell
def _(incremental_mde, mnist, n_train):
    _ax = pymde.plot(incremental_mde.X[:n_train], color_by=mnist.attributes['digits'][:n_train])
    _ax.set_title('Old embedding vectors')
    return


if __name__ == "__main__":
    app.run()



"""Internal utilities."""

import functools
import numbers

import numpy as np
import scipy.sparse
import torch

from pymde.average_distortion import _project_gradient


_DEVICE = torch.device("cpu")
_NP_RNG = np.random.default_rng()


class SolverError(Exception):
    pass


def get_default_device():
    return str(_DEVICE)


def _canonical_device(device):
    if isinstance(device, str):
        device = torch.device(device)
    elif not isinstance(device, torch.device):
        raise ValueError("device must be a str or a torch.device object.")

    if device.type == "cuda" and device.index is None:
        device = torch.device("cuda", torch.cuda.current_device())
    return device


def set_default_device(device):
    global _DEVICE
    _DEVICE = _canonical_device(device)


def _module_device(module):
    data = list(module.buffers())
    if not data:
        return None
    device = str(data[0].device)
    if any(str(datum.device) != device for datum in data):
        return None
    return device


def _is_numeric(arg):
    return (
        isinstance(arg, numbers.Number)
        or isinstance(arg, np.ndarray)
        or isinstance(arg, np.matrix)
        or isinstance(arg, torch.Tensor)
    )


def to_tensor(args, device=None):
    """Convert an arg or sequence of args to torch Tensors"""
    singleton = not isinstance(args, (list, tuple))
    if singleton:
        args = [args]

    tensor_args = []
    for arg in args:
        if isinstance(arg, torch.Tensor):
            tensor_args.append(arg)
        elif _is_numeric(arg):
            if isinstance(arg, np.ndarray) and arg.dtype == np.float64:
                tensor_args.append(
                    torch.tensor(arg, dtype=torch.float32, device=device)
                )
            else:
                tensor_args.append(torch.tensor(arg, device=device))
        else:
            raise ValueError("Received non-numeric argument ", arg)
    return tensor_args[0] if singleton else tensor_args


def tensor_arguments(func):
    """Cast numeric args and kwargs of func to Tensors."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        tensor_args = to_tensor(args)
        tensor_kwargs = {}
        for key, arg in kwargs.items():
            if isinstance(arg, torch.Tensor):
                tensor_kwargs[key] = arg
            elif _is_numeric(arg):
                tensor_kwargs[key] = torch.tensor(arg, device=_DEVICE)
            else:
                raise ValueError(
                    "Received non-numeric argument (name %s, value %s)"
                    % (key, arg)
                )
        return func(*tensor_args, **tensor_kwargs)

    return wrapper


def all_edges(n):
    """Return a tensor of all (n choose 2) edges

    Constructs all possible edges among n items. For example, if ``n`` is 4,
    the return value will be equal to

    .. code:: python3

        torch.Tensor([[0, 1], [0, 2], [0, 3], [1, 2], [1, 3], [2, 3]])
    """
    return torch.triu_indices(n, n, 1).T


@tensor_arguments
def natural_length(n, m):
    return (2.0 * n * m / (n - 1)).sqrt()


def in_stdemb(X):
    cov = (1.0 / X.shape[0]) * X.T @ X
    eye = torch.eye(2, dtype=X.dtype, device=X.device)
    mean = X.mean(axis=0)
    zero = torch.tensor(0.0, dtype=X.dtype, device=X.device)
    return torch.isclose(cov, eye).all() and torch.isclose(mean, zero).all()


def proj_standardized(X, demean=False, inplace=False):
    if demean:
        if inplace:
            X.sub_(X.mean(axis=0))
        else:
            X -= X.mean(axis=0)

    # pytorch 1.8.0 has a bug in which torch.svd fails when requires_grad
    # is true on the input (even if called under torch.no_grad)
    requires_grad = X.requires_grad
    X.requires_grad_(False)

    n = torch.tensor(X.shape[0], dtype=X.dtype, device=X.device)
    m = X.shape[1]
    # TODO: Gracefully handle the rare svd failure
    # TODO: debug alternative eigenvec approach ...
    # (evals, V = torch.eig(X.T @ X, eigenvectors=True)
    # proj = X @ V @ torch.diag(evals[:, 0].sqrt().pow(-1)) ...
    # proj *= torch.sqrt(n)
    if inplace:
        s = torch.zeros(m, device=X.device, dtype=X.dtype)
        Vh = torch.zeros((m, m), device=X.device, dtype=X.dtype)
        U = torch.zeros(X.shape, device=X.device, dtype=X.dtype)
        try:
            U, _, Vh = torch.linalg.svd(X, full_matrices=False, out=(U, s, Vh))
            V = Vh.transpose(-2, -1)
        except RuntimeError as e:
            X.requires_grad_(requires_grad)
            raise SolverError(str(e))
        torch.matmul(U[:, :m], V.T[:, :m], out=X)
        X.mul_(torch.sqrt(n))
        X.requires_grad_(requires_grad)
        return X
    else:
        try:
            U, _, Vh = torch.linalg.svd(X, full_matrices=False)
            V = Vh.transpose(-2, -1)
        except RuntimeError as e:
            X.requires_grad_(requires_grad)
            raise SolverError(str(e))
        proj = torch.sqrt(n) * U[:, :m] @ V.T[:, :m]
        X.requires_grad_(requires_grad)
        return proj


def adjacency_matrix(n, m, edges, weights, use_scipy=True):

    if use_scipy:
        if isinstance(weights, torch.Tensor):
            weights = weights.detach().cpu().numpy()
        if isinstance(edges, torch.Tensor):
            edges = edges.detach().cpu().numpy()
        A = scipy.sparse.coo_matrix(
            (weights, (edges[:, 0], edges[:, 1])),
            shape=(n, n),
            dtype=np.float32,
        )
        A = A + A.T
        return A.tocoo()
    else:
        A = torch.sparse_coo_tensor(
            edges.transpose(0, 1),
            weights,
            size=(n, n),
            dtype=torch.float32,
            device=edges.device,
        )
        A = A + A.transpose(0, 1)
        return A


@tensor_arguments
def procrustes(X_source, X_target):
    """min |X_source Q - X_target|_F s.t. Q^TQ = I"""
    U, _, Vh = torch.linalg.svd(X_target.T @ X_source, full_matrices=False)
    V = Vh.transpose(-2, -1)
    return V @ U.T


@tensor_arguments
def _rotate_2d(X, degrees):
    theta = torch.deg2rad(degrees)
    rot = torch.tensor(
        [
            [torch.cos(theta), -torch.sin(theta)],
            [torch.sin(theta), torch.cos(theta)],
        ],
        device=X.device,
    )
    return X @ rot


@tensor_arguments
def _rotate_3d(X, alpha, beta, gamma):
    alpha = torch.deg2rad(alpha.float())
    beta = torch.deg2rad(beta.float())
    gamma = torch.deg2rad(gamma.float())
    rot_x = torch.tensor(
        [
            [1, 0, 0],
            [0, torch.cos(alpha), torch.sin(alpha)],
            [0, -torch.sin(alpha), torch.cos(alpha)],
        ],
        device=X.device,
    )
    rot_y = torch.tensor(
        [
            [torch.cos(beta), 0.0, -torch.sin(beta)],
            [0, 1, 0],
            [torch.sin(beta), 0.0, torch.cos(beta)],
        ],
        device=X.device,
    )
    rot_z = torch.tensor(
        [
            [torch.cos(gamma), torch.sin(gamma), 0.0],
            [-torch.sin(gamma), torch.cos(gamma), 0.0],
            [0, 0, 1],
        ],
        device=X.device,
    )
    rot_3d = rot_x @ rot_y @ rot_z
    return X @ rot_3d


@tensor_arguments
def rotate(X, degrees):
    """Rotate a 2 or 3D embedding

    Rotates a 2/3D embedding by ``degrees``. If ``X`` is a 2D embedding,
    ``degrees`` should be a scalar; if it is 3D, ``degrees`` should be
    a length-3 ``torch.Tensor``, with one angle for each axis (the embedding
    will be rotated along the x-axis first, then the y-axis, then the z-axis).

    Arguments
    ---------
    X : torch.Tensor
        The embedding to rotate.
    degrees: torch.Tensor
        The angles of rotation.

    Returns
    -------
    torch.Tensor
        The rotated embedding
    """
    if X.shape[1] not in [2, 3]:
        raise ValueError(
            "Only 2 or 3 dimensional embeddings can be "
            "rotated using this method."
        )

    if X.shape[1] == 2:
        if degrees.numel() != 1:
            raise ValueError("`degrees` must be a scalar.")
        return _rotate_2d(X, degrees)
    else:
        if degrees.numel() != 3:
            raise ValueError("`degrees` must be a length-3 tensor.")
        return _rotate_3d(X, degrees[0], degrees[1], degrees[2])


@tensor_arguments
def center(X):
    """Center an embedding

    Returns a new embedding, equal to the given embedding minus the mean
    of its rows.
    """
    return X - X.mean(dim=0)[None, :]


@tensor_arguments
def align(source, target):
    """Align an source embedding to a target embedding

    Align the source embedding to another target embedding, via
    rotation. The alignment is done by solving an
    orthogonal Procrustes problem.

    Arguments
    ---------
    source: torch.Tensor
        The embedding to be aligned.
    target: torch.Tensor
        The target embedding, of the same shape as source.

    Returns
    -------
    torch.Tensor
        The rotation of source best aligned to the target.
    """
    source_mean = source.mean(dim=0)
    source = source - source_mean[None, :]
    source_col_rms = source.norm(dim=0)
    source = source / source_col_rms[None, :]

    target = center(target)
    target = target / target.norm(dim=0)

    Q = procrustes(source, target)
    rotated = source @ Q
    return (rotated * source_col_rms[None, :]) + source_mean


@tensor_arguments
def scale_delta(delta, d_nat):
    # scale delta so RMS(delta) == d_nat
    N = delta.nelement()
    rms = torch.sqrt(1 / N * torch.sum(delta**2))
    return delta * d_nat / rms


class LinearOperator(object):
    def __init__(self, matvec, device):
        self._matvec = matvec
        self.device = device

    def matvec(self, vecs):
        return self._matvec(vecs)


def make_hvp(f, edges, X, constraint):
    X_shape = X.shape

    def avg_dist_flat(X_flat):
        X_reshaped = X_flat.view(X_shape)
        if constraint is not None:
            # a noop in the forward pass, but projects the gradient onto
            # the tangent space of the constraint in the backward pass
            X_reshaped = _project_gradient(X_reshaped, constraint)
        # using custom average distortion yields a zero for hvp, since
        # gradient graph is disconnected
        differences = X_reshaped[edges[:, 0]] - X_reshaped[edges[:, 1]]
        norms = differences.pow(2).sum(dim=1).sqrt()
        return f(norms).mean()

    X_flat = X.view(-1).detach()

    def hvp(vecs):
        vecs = torch.split(vecs, 1, dim=1)
        products = []
        for v in vecs:
            _, product = torch.autograd.functional.vhp(
                avg_dist_flat, X_flat, v.squeeze()
            )
            products.append(product)
        return torch.stack(products, dim=1)

    return hvp


def hutchpp(linear_operator, dimension, n_queries):
    A = linear_operator
    d = dimension
    m = n_queries
    S = torch.randn(d, m // 3, device=A.device)
    G = torch.randn(d, m // 3, device=A.device)
    Q, _ = torch.qr(A.matvec(S))
    proj = G - Q @ (Q.T @ G)
    return torch.trace(Q.T @ A.matvec(Q)) + (3.0 / m) * torch.trace(
        proj.T @ A.matvec(proj)
    )


def np_rng():
    return _NP_RNG


def seed(seed: int):
    """Set the random seed

    This function sets the random seed that PyMDE uses in various
    preprocessing methods.
    """
    torch.manual_seed(seed)
    # Set numpy's legacy (global) random state for backward compatibility
    np.random.seed(seed)
    global _NP_RNG
    _NP_RNG = np.random.default_rng(seed)


def random_edges(n, p, seed=0):
    randomstate = np.random.default_rng(seed)
    edge_idx = randomstate.choice(
        int(n * (n - 1) / 2), p, replace=False, shuffle=False
    )
    u = (
        n
        - 2
        - np.floor(np.sqrt(-8 * edge_idx + 4 * n * (n - 1) - 7) / 2.0 - 0.5)
    )
    v = edge_idx + u + 1 - n * (n - 1) / 2 + (n - u) * ((n - u) - 1) / 2
    return torch.tensor(np.stack([u, v], axis=1).astype(np.int64))


class Distortion(torch.autograd.Function):
    """Manual implementation of the average distortion gradient, for testing"""

    @staticmethod
    def forward(ctx, X, f, A, lhs, rhs):
        distances = A.T @ X
        norms = distances.norm(dim=1)

        with torch.enable_grad():
            X.requires_grad_(False)
            norms.requires_grad_(True)
            norms.grad = None
            distortion = f(norms).mean()
            distortion.backward()
            g = norms.grad / norms
        X.requires_grad_(True)
        D = g.diag()

        grad_E = A @ (D @ (A.T @ X))
        ctx.grad_E = grad_E
        return distortion

    def backward(ctx, grad_output):
        return ctx.grad_E * grad_output, None, None, None, None


_distortion = Distortion.apply



# /// script
# requires-python = ">=3.12"
# dependencies = [
#     "altair==6.0.0",
#     "marimo",
#     "matplotlib==3.10.8",
#     "pandas==3.0.0",
#     "pymde==0.2.3",
#     "torch==2.10.0",
#     "wigglystuff==0.2.27",
# ]
# ///

import marimo

__generated_with = "0.19.11"
app = marimo.App(
    width="medium",
    css_file="/usr/local/_marimo/custom.css",
    auto_download=["html"],
)

with app.setup:
    import matplotlib.pyplot as plt
    import pandas as pd
    import pymde
    import torch
    from wigglystuff import ChartSelect

    import marimo as mo

    mnist = pymde.datasets.MNIST()


@app.function
def compute_embedding(embedding_dim, constraint, algorithm):
    mo.output.append(
        mo.md("Your embedding is being computed ... hang tight!").callout(kind="warn")
    )
    device = "cuda" if torch.cuda.is_available() else "cpu"
    if algorithm == "log1p":
        X = pymde.preserve_neighbors(
            mnist.data,
            embedding_dim=embedding_dim,
            constraint=constraint,
            device=device,
            verbose=True,
        ).embed(verbose=True)
    elif algorithm == "umap":
        X = pymde.preserve_neighbors(
            mnist.data,
            embedding_dim=embedding_dim,
            constraint=pymde.Centered(),
            device=device,
            verbose=True,
        ).embed(verbose=True)
    elif algorithm == "pca":
        X = pymde.pca(mnist.data, 2)
    else:
        X = pymde.laplacian_embedding(mnist.data, embedding_dim=2).embed(verbose=True)
    mo.output.clear()
    return X


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    # MNIST

    This notebook shows how to use the function `pymde.preserve_neighbors` to produce embeddings that highlight the local structure of your data, using MNIST as a case study. In these embeddings similar digits are near each other, and dissimilar digits are not near each other.

    It also shows how to debug embeddings and search for outliers in your original data, using PyMDE.
    """)
    return


@app.cell
def _():
    algorithm = mo.ui.dropdown(["log1p", "pca", "laplacian", "umap"], value="log1p")
    algorithm
    return (algorithm,)


@app.cell
def _(algorithm, constraint, embedding_dimension):
    with mo.persistent_cache("embedding"):
        embedding = compute_embedding(embedding_dimension, constraint, algorithm.value)

    embedding = pymde.rotate(embedding, 40) * torch.tensor([-1, -1])
    return (embedding,)


@app.cell(hide_code=True)
def _():
    mo.md("""
    Here's an **embedding of MNIST**: each point represents a digit,
    with similar digits close to each other.

    **Try making a selection with your mouse!**
    """)
    return


@app.cell
def _(embedding):
    ax = pymde.plot(embedding, color_by=mnist.attributes["digits"])
    plt.tight_layout()
    fig = mo.ui.anywidget(ChartSelect(ax.figure))
    fig
    return (fig,)


@app.cell
def _(embedding, fig):
    selected_indices = fig.get_indices(embedding[:, 0], embedding[:, 1])
    return (selected_indices,)


@app.cell(hide_code=True)
def _(selected_indices, table):
    # mo.stop() prevents this cell from running if the chart has
    # no selection
    mo.stop(not selected_indices.size)

    # show 10 images: either the first 10 from the selection, or the first ten
    # selected in the table
    selected_images = (
        show_images(list(selected_indices))
        if not len(table.value)
        else show_images(list(table.value["index"]))
    )

    mo.md(
        f"""
        **Here's a preview of the images you've selected**:

        {mo.as_html(selected_images)}

        Here's all the data you've selected.

        {table}
        """
    )
    return


@app.cell
def _(df, selected_indices):
    table = mo.ui.table(df.iloc[selected_indices])
    return (table,)


@app.cell
def _():
    embedding_dimension = 2
    constraint = pymde.Standardized()
    return constraint, embedding_dimension


@app.cell
def _(embedding):
    indices = torch.arange(mnist.data.shape[0]).numpy()

    df = pd.DataFrame(
        {
            "index": indices,
            "x": embedding[:, 0],
            "y": embedding[:, 1],
            "digit": mnist.attributes["digits"][indices],
        }
    )
    return (df,)


@app.function
def show_images(indices, max_images=10):
    indices = indices[:max_images]
    images = mnist.data.reshape((-1, 28, 28))[indices]
    fig, axes = plt.subplots(1, len(indices))
    fig.set_size_inches(12.5, 1.5)
    if len(indices) > 1:
        for im, ax in zip(images, axes.flat):
            ax.imshow(im, cmap="gray")
            ax.set_yticks([])
            ax.set_xticks([])
    else:
        axes.imshow(images[0], cmap="gray")
        axes.set_yticks([])
        axes.set_xticks([])
    plt.tight_layout()
    return fig


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## Embeddings, in one line
    """)
    return


@app.cell
def _():
    _embedding = pymde.preserve_neighbors(mnist.data).embed(verbose=True)
    pymde.plot(_embedding, color_by=mnist.attributes["digits"])
    return


@app.cell
def _():
    _embedding = pymde.preserve_neighbors(
        mnist.data, constraint=pymde.Standardized()
    ).embed(verbose=True)
    pymde.plot(_embedding, color_by=mnist.attributes["digits"])
    return


@app.cell
def _():
    _embedding = pymde.preserve_neighbors(
        mnist.data,
        attractive_penalty=pymde.penalties.Quadratic,
        repulsive_penalty=None,
    ).embed(verbose=True)
    pymde.plot(_embedding, color_by=mnist.attributes["digits"])
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## Embeddings, from scratch
    """)
    return


@app.cell
def _():
    knn_graph = pymde.preprocess.k_nearest_neighbors(
        mnist.data, k=15, verbose=True
    )
    return (knn_graph,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ### Quadratic MDE problems
    """)
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    The `embed` method computes and returns an embedding.

    The embedding is also accessible via the `X` attribute of the `MDE` instance.
    """)
    return


@app.cell
def _(knn_graph):
    quadratic_mde = pymde.MDE(
        n_items=mnist.data.shape[0],
        embedding_dim=2,
        edges=knn_graph.edges,
        distortion_function=pymde.penalties.Quadratic(knn_graph.weights),
        constraint=pymde.Standardized(),
    )

    quadratic_mde.embed(verbose=True)
    return (quadratic_mde,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    After embedding, we can call the `plot` method to visualize the embedding (when `embedding_dim` $\leq 3$).

    The `color_by` keyword argument takes a length-$n$ list of attribute values associated with the items; the values are used to color the points, with each unique value getting its own color.

    Here, the attribute is the digit depicted by the image.
    """)
    return


@app.cell
def _(quadratic_mde):
    quadratic_mde.plot(color_by=mnist.attributes["digits"])
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    Next, we fit an embedding with 3 dimensions. We can visualize 3D embeddings with `PyMDE`.
    """)
    return


@app.cell
def _(knn_graph):
    quadratic_mde_3d = pymde.MDE(
        n_items=mnist.data.shape[0],
        embedding_dim=3,
        edges=knn_graph.edges,
        distortion_function=pymde.penalties.Quadratic(knn_graph.weights),
        constraint=pymde.Standardized(),
    )

    quadratic_mde_3d.embed(verbose=True)
    quadratic_mde_3d.plot(color_by=mnist.attributes["digits"])
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## Debugging the embedding
    """)
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ### The distribution of distortions

    We can visualize the cumulative distribution of distortions. We see that a small number of pairs are responsible for most of the embedding's distortion.
    """)
    return


@app.cell
def _(quadratic_mde):
    quadratic_mde.distortions_cdf()
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ### High distortion pairs

    It can be instructive to inspect the pairs with highest distortion. A pair, here one column of the grid, is a set of two images that our MDE problem was told are similar. In this case, some of the high-distortion pairs are oddly drawn digits, and don't appear very similar to each other.
    """)
    return


@app.cell
def _(quadratic_mde):
    pairs, distortions = quadratic_mde.high_distortion_pairs()
    outliers = pairs[:10]


    def plot_pairs(pairs):
        fig, axs = plt.subplots(2, pairs.shape[0], figsize=(15.0, 3.0))
        for pair_index in range(pairs.shape[0]):
            i = pairs[pair_index][0]
            j = pairs[pair_index][1]
            im_i = mnist.data[i].reshape(28, 28)
            im_j = mnist.data[j].reshape(28, 28)
            axs[0][pair_index].imshow(im_i)
            axs[0][pair_index].set_xticks([])
            axs[0][pair_index].set_yticks([])
            axs[1][pair_index].imshow(im_j)
            axs[1][pair_index].set_xticks([])
            axs[1][pair_index].set_yticks([])
        plt.tight_layout()


    plot_pairs(outliers)
    return pairs, plot_pairs


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## Low distortion pairs

    In comparison, the pairs with the lowest distortion look like very reasonable pairs of images.
    """)
    return


@app.cell
def _(pairs, plot_pairs):
    low_distortion_pairs = pairs[-10:]
    plot_pairs(low_distortion_pairs)
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## Including dissimilar pairs

    The quadratic MDE problem used a standardization constraint to enforce the embedding to spread out, and only used pairs of similar items.

    Instead of only relying on the constraint, we can also design the distortion functions so that they discourage dissimilar items from being close. We do this by including some pairs of dissimilar items, in addition to the pairs of similar items.
    """)
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ### Form $\mathcal{E} = \mathcal{E}_\text{sim} \cup \mathcal{E}_\text{dis}$
    """)
    return


@app.cell
def _(knn_graph):
    similar_edges = knn_graph.edges

    dissimilar_edges = pymde.preprocess.dissimilar_edges(
        n_items=mnist.data.shape[0],
        num_edges=similar_edges.shape[0],
        similar_edges=similar_edges,
    )

    edges = torch.cat([similar_edges, dissimilar_edges])
    weights = torch.cat(
        [knn_graph.weights, -1.0 * torch.ones(dissimilar_edges.shape[0])]
    )

    plt.hist(weights.numpy())
    plt.gca()
    return edges, weights


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ### Define the distortion function $f$
    """)
    return


@app.cell
def _(weights):
    f = pymde.penalties.PushAndPull(
        weights=weights,
        attractive_penalty=pymde.penalties.Log1p,
        repulsive_penalty=pymde.penalties.Log,
    )
    return (f,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ### A standardized embedding ...
    """)
    return


@app.cell
def _(edges, f, quadratic_mde):
    std_mde = pymde.MDE(
        n_items=mnist.data.shape[0],
        embedding_dim=2,
        edges=edges,
        distortion_function=f,
        constraint=pymde.Standardized(),
    )
    std_mde.embed(X=quadratic_mde.X, max_iter=400, verbose=True)
    std_mde.plot(color_by=mnist.attributes["digits"])
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## ... and an unconstrained embedding
    """)
    return


@app.cell
def _(edges, f, quadratic_mde):
    unconstrained_mde = pymde.MDE(
        n_items=mnist.data.shape[0],
        embedding_dim=2,
        edges=edges,
        distortion_function=f,
    )
    unconstrained_mde.embed(X=quadratic_mde.X, verbose=True)
    unconstrained_mde.plot(color_by=mnist.attributes["digits"])
    return


if __name__ == "__main__":
    app.run()



# /// script
# requires-python = "==3.13"
# dependencies = [
#     "marimo",
#     "matplotlib==3.10.8",
#     "numpy==2.3.5",
#     "openai==2.20.0",
#     "pandas==3.0.0",
#     "pydantic-ai-slim==1.58.0",
#     "pymde==0.2.3",
#     "torch==2.10.0",
#     "wigglystuff==0.2.22",
# ]
# ///

import marimo

__generated_with = "0.19.11"
app = marimo.App(
    width="medium",
    css_file="/usr/local/_marimo/custom.css",
    auto_download=["html"],
)

with app.setup:
    import marimo as mo
    import pymde
    import matplotlib.pyplot as plt
    import numpy as np
    import torch

    from matplotlib import colors
    from wigglystuff import ChartSelect

    DEVICE = "cuda" if torch.cuda.is_available() else "cpu"


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    # Google Scholar

    This notebook shows how to use the function `pymde.preserve_distances` to produce embeddings of networks, in which the goal is to preserve the shortest-path distances in the network.

    It uses an academic co-authorship network collected from Google Scholar as a case study.
    """)
    return


@app.cell
def _():
    gscholar = pymde.datasets.google_scholar()
    return (gscholar,)


@app.cell
def _(gscholar):
    scholars_df = gscholar.other_data['dataframe']
    scholars_df
    return (scholars_df,)


@app.cell
def _(gscholar):
    coauthorship_graph = gscholar.data
    return (coauthorship_graph,)


@app.cell
def _(coauthorship_graph):
    mo.md(f"""
    {coauthorship_graph.n_items:,} authors
    """)
    return


@app.cell
def _(coauthorship_graph):
    mo.md(f"""
    {coauthorship_graph.n_edges:,} edges
    """)
    return


@app.cell
def _(coauthorship_graph):
    mo.md(f"""
    edge density: {100*(coauthorship_graph.n_edges / (coauthorship_graph.n_all_edges)):.2f} percent
    """)
    return


@app.cell
def _(coauthorship_graph):
    with mo.persistent_cache("distance_preserving_mde"):
        distance_preserving_mde = pymde.preserve_distances(
            data=coauthorship_graph,
            loss=pymde.losses.Absolute,
            max_distances=100000000.0,
            device=DEVICE,
            verbose=True,
        )
    return (distance_preserving_mde,)


@app.cell
def _(distance_preserving_mde):
    plt.figure(figsize=(12, 3))
    original_distances = np.sort(distance_preserving_mde.distortion_function.deviations.cpu().numpy())
    ax = plt.gca()
    plt.hist(original_distances, histtype='step', bins=np.arange(1, 11), density=True, cumulative=True)
    plt.xlim(1, 10)
    plt.xticks(np.arange(1, 11))
    plt.xlabel('graph distances')
    plt.gca()
    return


@app.cell
def _(distance_preserving_mde):
    with mo.persistent_cache("solved_scholar"):
        distance_preserving_mde.embed(verbose=True)
        solved_mde = distance_preserving_mde
    return (solved_mde,)


@app.cell
def _(solved_mde):
    solved_mde.distortions_cdf()
    plt.gca()
    return


@app.cell
def _(gscholar, solved_mde):
    solved_mde.plot(
        color_by=gscholar.attributes["coauthors"],
        color_map="viridis",
        figsize_inches=(12.0, 12.0),
        background_color="k",
    )
    fig = plt.gcf()
    return (fig,)


@app.cell
def _(coauthorship_graph, gscholar, solved_mde):
    edges = coauthorship_graph.edges
    indices = torch.randperm(edges.shape[0])[:1000]
    edges = edges[indices].cpu().numpy()

    solved_mde.plot(
        edges=edges,
        color_by=gscholar.attributes["coauthors"],
        color_map="viridis",
        figsize_inches=(12, 12),
    )
    return


@app.cell
def _(gscholar):
    legend = {
        "bio": colors.to_rgba("tab:purple"),
        "ai": colors.to_rgba("tab:red"),
        "cs": colors.to_rgba("tab:cyan"),
        "ee": colors.to_rgba("tab:green"),
        "physics": colors.to_rgba("tab:orange"),
    }
    scholar_disciplines_df = gscholar.other_data["disciplines"]
    topic_colors = [legend[code] for code in scholar_disciplines_df["topic"]]
    return scholar_disciplines_df, topic_colors


@app.cell
def _(scholar_disciplines_df, solved_mde, topic_colors):
    pymde.plot(
        solved_mde.X[scholar_disciplines_df["node_id"].values],
        colors=topic_colors,
        figsize_inches=(12, 12),
        background_color="black",
    )
    return


@app.cell
def _(fig):
    select = mo.ui.anywidget(ChartSelect(fig))
    select
    return (select,)


@app.cell
def _(scholars_df, select, solved_mde):
    _X = solved_mde.X.cpu()
    mask = select.get_mask(_X[:, 0], _X[:, 1])
    scholars_df.iloc[mask]
    return


if __name__ == "__main__":
    app.run()



import math
import time

import torch

from pymde.average_distortion import _average_distortion, _gather_indices
from pymde.lbfgs import LBFGS
from pymde import util


class SolveStats(object):
    """Summary statistics for a solve.

    Attributes
    ----------
    average_distortions: sequence
        The average distortion at each iteration.
    residual_norms: sequence
        The residual norm at each iteration.
    step_size_percents: sequence
        The relative size of each step.
    solve_time: float
        The time it took to create the embedding, in seconds.
    snapshots: sequence
        Snapshots of the embedding.
    snapshot_every: int
        The number of iterations between snapshots.
    """

    def __init__(
        self,
        average_distortions,
        residual_norms,
        step_size_percents,
        solve_time,
        times,
        snapshots,
        snapshot_every,
    ):
        self.average_distortions = average_distortions
        self.residual_norms = residual_norms
        self.step_size_percents = step_size_percents
        self.solve_time = solve_time
        self.iterations = len(average_distortions)
        self.times = times
        self.snapshots = snapshots
        self.snapshot_every = snapshot_every

    def __str__(self):
        return (
            "SolveStats:\n"
            "\taverage distortion {0:.3g}\n"
            "\tresidual norm {1:.3g}\n"
            "\tsolve_time (s) {2:.3g}\n"
            "\titerations {3}".format(
                self.average_distortions[-1],
                self.residual_norms[-1],
                self.solve_time,
                self.iterations,
            )
        )

    def _repr_pretty_(self, p, cycle):
        del cycle
        text = self.__str__()
        p.text(text)


def lbfgs(
    X,
    objective_fn,
    constraint,
    eps,
    max_iter,
    memory_size,
    use_line_search,
    use_cached_loss,
    verbose,
    print_every,
    snapshot_every,
    logger,
):
    start_time = time.time()
    average_distortions = []
    grad_norms = []
    step_size_percents = []
    times = []
    snapshots = []

    # LBFGS set-up
    #
    # This callback logs the average distortion and gradient norm
    # per teration
    def callback(loss, grad):
        average_distortions.append(loss.detach().cpu().item())
        grad_norms.append(grad.detach().norm(p="fro").cpu().item())

    # closure for LBFGS optimization step; projects gradient onto
    # onto the tangent space of the iterate
    def value_and_grad():
        opt.zero_grad()
        average_distortion = objective_fn(X)
        average_distortion.backward()
        constraint.project_onto_tangent_space(X, X.grad, inplace=True)
        return average_distortion

    ls = "strong_wolfe" if use_line_search else None
    opt = LBFGS(
        params=[X],
        max_iter=1,
        history_size=memory_size,
        line_search_fn=ls,
        callback=callback,
        tolerance_grad=-torch.tensor(
            float("inf"), dtype=X.dtype, device=X.device
        ),
        tolerance_change=-torch.tensor(
            float("inf"), dtype=X.dtype, device=X.device
        ),
        project_callback=constraint.project_onto_constraint,
        use_cached_loss=use_cached_loss,
    )

    digits = len(str(max_iter))
    start = time.time()
    for iteration in range(0, max_iter):
        if snapshot_every is not None and iteration % snapshot_every == 0:
            snapshots.append(X.detach().cpu().clone())
        with torch.no_grad():
            norm_X = X.norm(p="fro")
        X.requires_grad_(True)
        opt.step(value_and_grad)
        X.requires_grad_(False)

        with torch.no_grad():
            constraint.project_onto_constraint(X, inplace=True)

        times.append(time.time() - start)
        try:
            h = opt.state[opt._params[0]]["t"]
            d = opt.state[opt._params[0]]["d"]
            percent_change = 100 * h * d.norm() / norm_X
        except KeyError:
            h = 0.0
            d = 0.0
            percent_change = 0.0
        step_size_percents.append(float(percent_change))

        norm_grad = grad_norms[-1]
        if verbose and (
            (iteration % print_every == 0) or (iteration == max_iter - 1)
        ):
            logger.info(
                "iteration %0*d | distortion %6f | residual norm %g | "
                "step length %g | percent change %g"
                % (
                    digits,
                    iteration,
                    average_distortions[-1],
                    norm_grad,
                    h,
                    percent_change,
                )
            )
        if norm_grad <= eps:
            if verbose:
                logger.info(
                    "Converged in %03d iterations, with residual norm %g"
                    % (iteration + 1, norm_grad)
                )
            break
        elif h == 0:
            opt.reset()
    tot_time = time.time() - start_time
    solve_stats = SolveStats(
        average_distortions,
        grad_norms,
        step_size_percents,
        tot_time,
        times,
        snapshots,
        snapshot_every,
    )
    return X, solve_stats


def spi(
    batch_size,
    stochastic_function,
    X,
    constraint,
    eps,
    max_iter,
    memory_size,
    use_line_search,
    verbose,
    print_every,
    snapshot_every,
    logger,
):
    average_distortions = []
    step_size_percents = []
    times = []
    snapshots = []

    X_target = torch.empty(X.shape, dtype=X.dtype, device=X.device)
    X_curr = X
    X_curr.requires_grad_(False)
    del X

    start_time = time.time()
    for iteration in range(0, max_iter):
        if snapshot_every is not None and iteration % snapshot_every == 0:
            snapshots.append(X_curr.detach().cpu().clone())
        f_batch, edges_batch = stochastic_function.sample(batch_size)
        lhs = _gather_indices(edges_batch[:, 0], X_curr.shape[1])
        rhs = _gather_indices(edges_batch[:, 1], X_curr.shape[1])
        average_distortion_batch = lambda X: _average_distortion(
            X, f_batch, lhs, rhs
        )

        X_target[:] = X_curr

        def solve_batch(X, fn, max_iter):
            try:
                _, solve_stats = lbfgs(
                    X=X,
                    objective_fn=fn,
                    constraint=constraint,
                    eps=eps,
                    max_iter=max_iter,
                    memory_size=memory_size,
                    use_line_search=use_line_search,
                    use_cached_loss=False,
                    verbose=False,
                    print_every=print_every,
                    snapshot_every=None,
                    logger=logger,
                )
            except util.SolverError as e:
                # NaN encountered, svd failed ...
                # skip this iteration and hope that the next batch of edges is
                # okay ...
                logger.warning("error in batch: %s" % str(e))
                solve_stats = None
            X.requires_grad_(False)
            return solve_stats

        if iteration == 0:
            objective_fn = average_distortion_batch
        else:

            def objective_fn(X):
                avg_dist = average_distortion_batch(X)
                prox_term = c * iteration / 2.0 * (X - X_target).pow(2).sum()
                return avg_dist + prox_term

        start = time.time()
        solve_stats = solve_batch(X_curr, objective_fn, 20)
        times.append(time.time() - start)

        if torch.isnan(X_curr).any():
            X_curr[:] = X_target
            break

        with torch.no_grad():
            average_distortions.append(
                average_distortion_batch(X_curr).cpu().item()
            )
            dX = X_curr - X_target
            step_size_percents.append(
                (100 * dX.norm() / X_target.norm()).cpu().item()
            )
            del dX

        if verbose and iteration % print_every == 0:
            iterations = 0 if solve_stats is None else solve_stats.iterations
            logger.info(
                "round %d | distortion %6f | "
                "percent change %g | LBFGS iterations %d"
                % (
                    iteration,
                    average_distortions[-1],
                    step_size_percents[-1],
                    iterations,
                )
            )

        if iteration == 0:
            del X_target
            hvp = util.make_hvp(f_batch, edges_batch, X_curr, constraint)
            linop = util.LinearOperator(hvp, device=X_curr.device)
            trace_estimate = util.hutchpp(linop, X_curr.nelement(), 10)
            c = trace_estimate / (X_curr.nelement())
            c /= 10.0
            if verbose:
                logger.info("Using c = %f" % c)
            X_target = X_curr.detach().clone()

    tot_time = time.time() - start_time
    return (
        X_curr,
        SolveStats(
            average_distortions,
            [math.inf],
            step_size_percents,
            tot_time,
            times,
            snapshots,
            snapshot_every,
        ),
    )



"""Benchmark k-nearest neighbor methods.

Usage:
    python benchmarks/bench_knn.py [--output FILE] [--skip-pynndescent]
                                   [--sizes ...] [--dims ...] [--k K]
"""

import argparse
import csv
import io
import sys
import time
import tracemalloc

import numpy as np


def _ground_truth(data, k):
    """Brute-force exact kNN using Rust extension for recall computation."""
    from pymde._native import knn_l2

    neighbors, _ = knn_l2(data, k)
    return neighbors[:, 1:]  # drop self


def recall_at_k(neighbors, gt_neighbors):
    """Fraction of true neighbors found, averaged over all queries."""
    n, k = gt_neighbors.shape
    neighbors = neighbors[:, :k]
    hits = 0
    for i in range(n):
        hits += len(set(neighbors[i]) & set(gt_neighbors[i]))
    return hits / (n * k)


def bench_rust_knn(data, k):
    from pymde._native import knn_l2

    tracemalloc.start()
    t0 = time.perf_counter()
    neighbors, sq_distances = knn_l2(data, k)
    elapsed = time.perf_counter() - t0
    _, peak_mem = tracemalloc.get_traced_memory()
    tracemalloc.stop()
    distances = np.sqrt(np.maximum(sq_distances[:, 1:], 0.0))
    return neighbors[:, 1:], distances, elapsed, peak_mem


def bench_rust_nndescent(data, k):
    from pymde._native import nn_descent

    tracemalloc.start()
    t0 = time.perf_counter()
    neighbors, distances = nn_descent(data, k)
    elapsed = time.perf_counter() - t0
    _, peak_mem = tracemalloc.get_traced_memory()
    tracemalloc.stop()
    return neighbors[:, 1:], distances[:, 1:], elapsed, peak_mem


def bench_pynndescent(data, k):
    import pynndescent

    tracemalloc.start()
    t0 = time.perf_counter()
    index = pynndescent.NNDescent(
        data, n_neighbors=k + 1, verbose=False, max_candidates=60
    )
    neighbors, distances = index.neighbor_graph
    elapsed = time.perf_counter() - t0
    _, peak_mem = tracemalloc.get_traced_memory()
    tracemalloc.stop()
    return neighbors[:, 1:], distances[:, 1:], elapsed, peak_mem


def bench_instant_distance(data, k):
    import instant_distance

    n, d = data.shape
    tracemalloc.start()
    t0 = time.perf_counter()

    config = instant_distance.Config()
    config.ef_construction = 200
    config.ef_search = 200
    points = [data[i].tolist() for i in range(n)]
    hnsw, ids = instant_distance.Hnsw.build(points, config)

    # ids[i] = internal pid for original point i; build reverse map
    pid_to_idx = {pid: idx for idx, pid in enumerate(ids)}

    all_neighbors = np.empty((n, k), dtype=np.int64)
    all_distances = np.empty((n, k), dtype=np.float32)
    search = instant_distance.Search()

    for i in range(n):
        hnsw.search(points[i], search)
        found = 0
        for neighbor in search:
            idx = pid_to_idx[neighbor.pid]
            if idx == i:
                continue
            if found < k:
                all_neighbors[i, found] = idx
                all_distances[i, found] = neighbor.distance
                found += 1
            if found >= k:
                break
        for j in range(found, k):
            all_neighbors[i, j] = -1
            all_distances[i, j] = float("inf")

    elapsed = time.perf_counter() - t0
    _, peak_mem = tracemalloc.get_traced_memory()
    tracemalloc.stop()
    return all_neighbors, all_distances, elapsed, peak_mem


def bench_faiss_flat(data, k):
    import faiss

    d = data.shape[1]
    tracemalloc.start()
    t0 = time.perf_counter()
    index = faiss.IndexFlatL2(d)
    index.add(data)
    sq_distances, neighbors = index.search(data, k + 1)
    elapsed = time.perf_counter() - t0
    _, peak_mem = tracemalloc.get_traced_memory()
    tracemalloc.stop()
    distances = np.sqrt(np.maximum(sq_distances[:, 1:], 0.0))
    return neighbors[:, 1:], distances, elapsed, peak_mem


def bench_faiss_knn(data, k):
    import faiss
    from faiss.contrib.exhaustive_search import knn

    tracemalloc.start()
    t0 = time.perf_counter()
    sq_distances, neighbors = knn(data, data, k + 1)
    elapsed = time.perf_counter() - t0
    _, peak_mem = tracemalloc.get_traced_memory()
    tracemalloc.stop()
    distances = np.sqrt(np.maximum(sq_distances[:, 1:], 0.0))
    return neighbors[:, 1:], distances, elapsed, peak_mem


def bench_faiss_ivf(data, k):
    import faiss

    n, d = data.shape
    nlist = max(1, int(np.sqrt(n)))
    tracemalloc.start()
    t0 = time.perf_counter()
    quantizer = faiss.IndexFlatL2(d)
    index = faiss.IndexIVFFlat(quantizer, d, nlist)
    index.train(data)
    index.add(data)
    index.nprobe = min(10, nlist)
    sq_distances, neighbors = index.search(data, k + 1)
    elapsed = time.perf_counter() - t0
    _, peak_mem = tracemalloc.get_traced_memory()
    tracemalloc.stop()
    distances = np.sqrt(np.maximum(sq_distances[:, 1:], 0.0))
    return neighbors[:, 1:], distances, elapsed, peak_mem


# Always-available methods
METHODS = {
    "rust-knn": bench_rust_knn,
    "rust-nndescent": bench_rust_nndescent,
}

# Conditionally add faiss methods
try:
    import faiss  # noqa: F401

    METHODS["faiss-flat"] = bench_faiss_flat
    METHODS["faiss-knn"] = bench_faiss_knn
    METHODS["faiss-ivf"] = bench_faiss_ivf
except ImportError:
    pass

try:
    import pynndescent

    METHODS["pynndescent"] = bench_pynndescent
except ImportError:
    pass

try:
    import instant_distance  # noqa: F401

    METHODS["instant-distance"] = bench_instant_distance
except ImportError:
    pass


def run_benchmarks(sizes, dims, k, skip_pynndescent=False):
    methods = {
        name: fn
        for name, fn in METHODS.items()
        if not (skip_pynndescent and name == "pynndescent")
    }

    results = []
    for n in sizes:
        for d in dims:
            print(f"\n--- n={n}, d={d}, k={k} ---")
            rng = np.random.default_rng(42)
            data = rng.standard_normal((n, d)).astype(np.float32)

            # Compute ground truth once
            gt_neighbors = _ground_truth(data, k)

            for name, fn in methods.items():
                print(f"  {name:20s} ... ", end="", flush=True)
                try:
                    neighbors, distances, elapsed, peak_mem = fn(data, k)
                    rec = recall_at_k(neighbors, gt_neighbors)
                    peak_mb = peak_mem / (1024 * 1024)
                    print(
                        f"time={elapsed:7.2f}s  "
                        f"mem={peak_mb:7.1f}MB  "
                        f"recall={rec:.4f}"
                    )
                    results.append(
                        {
                            "n": n,
                            "d": d,
                            "k": k,
                            "method": name,
                            "time_s": round(elapsed, 4),
                            "peak_mem_mb": round(peak_mb, 2),
                            "recall": round(rec, 6),
                        }
                    )
                except Exception as e:
                    print(f"FAILED: {e}")
                    results.append(
                        {
                            "n": n,
                            "d": d,
                            "k": k,
                            "method": name,
                            "time_s": None,
                            "peak_mem_mb": None,
                            "recall": None,
                        }
                    )
    return results


def format_table(results):
    if not results:
        return ""
    buf = io.StringIO()
    header = f"{'method':20s} {'n':>8s} {'d':>5s} {'k':>4s} {'time(s)':>9s} {'mem(MB)':>9s} {'recall':>8s}"
    buf.write(header + "\n")
    buf.write("-" * len(header) + "\n")
    for r in results:
        t = f"{r['time_s']:9.2f}" if r["time_s"] is not None else "    FAIL"
        m = (
            f"{r['peak_mem_mb']:9.1f}"
            if r["peak_mem_mb"] is not None
            else "    FAIL"
        )
        rec = f"{r['recall']:8.4f}" if r["recall"] is not None else "    FAIL"
        buf.write(
            f"{r['method']:20s} {r['n']:8d} {r['d']:5d} {r['k']:4d} {t} {m} {rec}\n"
        )
    return buf.getvalue()


def main():
    parser = argparse.ArgumentParser(description="Benchmark kNN methods")
    parser.add_argument(
        "--output", type=str, default=None, help="CSV output file"
    )
    parser.add_argument(
        "--skip-pynndescent", action="store_true", help="Skip pynndescent"
    )
    parser.add_argument(
        "--sizes",
        type=int,
        nargs="+",
        default=[1000, 10000, 50000, 100000, 500000],
    )
    parser.add_argument("--dims", type=int, nargs="+", default=[32, 128, 784])
    parser.add_argument("--k", type=int, default=15)
    args = parser.parse_args()

    results = run_benchmarks(
        args.sizes, args.dims, args.k, skip_pynndescent=args.skip_pynndescent
    )

    print("\n\n=== Summary ===\n")
    print(format_table(results))

    if args.output:
        fieldnames = [
            "method",
            "n",
            "d",
            "k",
            "time_s",
            "peak_mem_mb",
            "recall",
        ]
        with open(args.output, "w", newline="") as f:
            writer = csv.DictWriter(f, fieldnames=fieldnames)
            writer.writeheader()
            writer.writerows(results)
        print(f"\nResults written to {args.output}")


if __name__ == "__main__":
    main()



# /// script
# requires-python = ">=3.10"
# dependencies = [
#     "marimo",
#     "pymde",
#     "torch",
# ]
# ///

import marimo

__generated_with = "0.19.11"
app = marimo.App()

with app.setup:
    import marimo as mo
    import matplotlib.pyplot as plt
    import pymde
    import torch


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    # Single-Cell Genomics

    In this example notebook, we show how to use PyMDE to compute an embedding of some single-cell genomics data.

    We'll embed roughly 44,000 PBMCs, specified by their mRNA transcriptomes, into 3 dimensions. These PBMCs were collected from some patients who were healthy, and others who had severe COVID-19 infections.

    After embedding, we'll plot the embedding and analyze it by coloring the vectors by cell type and patient health status. We'll also analyze where the embedding succeeded, and where it failed.
    """)
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ### The original data
    The data is from [A single-cell atlas of the peripheral immune response in patients with severe COVID-19](https://www.nature.com/articles/s41591-020-0944-y), by Wilk, et al (2020).

    The original data is available at https://cellxgene.cziscience.com/
    """)
    return


@app.cell
def _():
    scrna_wilk = pymde.datasets.covid19_scrna_wilk()
    return (scrna_wilk,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    The data is a matrix in which each row is a PCA embedding of a cell's mRNA transcriptome.
    """)
    return


@app.cell
def _(scrna_wilk):
    scrna_wilk.data.shape
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    We can visualize the top 3 components of the PCA embedding.
    """)
    return


@app.cell
def _(scrna_wilk):
    pymde.plot(scrna_wilk.data, color_by=scrna_wilk.attributes['cell_type'], color_map='tab20')
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ### Embedding

    We'll compute an embedding via `pymde.preserve_neighbors`. We use this function, instead of `pymde.preserve_distances`, because we're interested more in the local structure of the data than the global.
    The exact distance between transcriptomes does not matter much, and might not even be that reliable.

    In the below cell, we've chosen to compute a standardized embedding. Standardized embeddings are generally reasonable: the constraint prevents the vectors from spreading out too much, and enforces the feature columns of the embedding to be uncorrelated.

    Feel free to experiment with the keyword arguments of the `preserve_neighbors` function. For example, you can omit the constraint (replace it with `None`), vary the `repulsive_fraction` or the number of neighbors `n_neighbors`, try a random initialization (`init='random'`), or modify the attractive and repulsive penalties.

    For a full list of the configurable options, type `pymde.preserve_neighbors?` into a code cell, and execute it.
    """)
    return


@app.cell
def _(scrna_wilk):
    mde = pymde.preserve_neighbors(
        scrna_wilk.data,
        embedding_dim=3,
        constraint=pymde.Standardized(),
        repulsive_fraction=1.0,
        verbose=True,
    )
    embedding = mde.embed(verbose=True)
    return embedding, mde


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    Next, we plot the embedding.

    Because the embedding is in 3 dimensions, some viewing angles look better than others. You can use `pymde.rotate` to rotate the embedding until you find a good angle.

    (Alternatively, use external libraries like Polyscope to explore 3D embeddings interactively.)
    """)
    return


@app.cell
def _(embedding, scrna_wilk):
    angles = torch.tensor([50., 167., 225.])
    rotated_embedding = pymde.rotate(embedding, angles)
    pymde.plot(rotated_embedding, color_by=scrna_wilk.attributes['cell_type'], color_map='tab20',
               figsize_inches=(12, 12))
    return (rotated_embedding,)


@app.cell
def _(rotated_embedding, scrna_wilk):
    pymde.plot(rotated_embedding, color_by=scrna_wilk.attributes['health_status'], color_map='coolwarm',
               figsize_inches=(12, 12))
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    Below, we plot a random sample of edges between pairs of similar cells. You'll see that the edges trace out the clusters in the embedding, and that different clusters are loosely connected to each other.
    """)
    return


@app.cell
def _(mde, rotated_embedding, scrna_wilk):
    similar_cells = mde.edges[mde.distortion_function.pos_idx]
    sampled_edges = similar_cells[torch.randperm(similar_cells.shape[0])[:10000]]
    pymde.plot(rotated_embedding, color_by=scrna_wilk.attributes['cell_type'], color_map='tab20', edges=sampled_edges, figsize_inches=(12, 12))
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    In the next cell, we plot the 500 pairs of cells that experienced the largest distortion. These represent pairs of cells that were difficult to embed.

    In general, it can be instructive to manually inspect the original data corresponding to the most highly distorted pairs. You might find that the pairs don't actually represent similar items (cells), or maybe that the items are in some way outliers, anomalous, or corrupted.
    """)
    return


@app.cell
def _(mde, rotated_embedding, scrna_wilk):
    pairs, _ = mde.high_distortion_pairs()
    most_distorted_pairs = pairs[:500]
    pymde.plot(rotated_embedding, color_by=scrna_wilk.attributes['cell_type'], color_map='tab20',
               edges=most_distorted_pairs, figsize_inches=(12, 12))
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    Finally, we plot the CDF of distortions below. You can see that most cells experience little distortions, but a few cells --- roughly 10 or 20 percent --- have much higher distortion.
    """)
    return


@app.cell
def _(mde):
    mde.distortions_cdf()
    plt.gca()
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    In this notebook, we computed a neighbor-preserving embedding, since we were more interested in the local structure than the global structure of the data.

    If you are curious what a distance-preserving embedding might look like, you can compute one using the following code:

    ```python3

    mde = pymde.preserve_distances(
        scrna_wilk.data,
        embedding_dim=3,
        verbose=True,
        max_distances=1e7,
    )
    embedding = mde.embed(verbose=True)
    ```

    You'll find that it looks similar to the original PCA embedding.
    """)
    return


if __name__ == "__main__":
    app.run()



# /// script
# requires-python = ">=3.10"
# dependencies = [
#     "marimo",
#     "matplotlib==3.10.8",
#     "pandas",
#     "pymde==0.2.3",
#     "torch==2.10.0",
#     "wigglystuff==0.2.27",
# ]
# ///

import marimo

__generated_with = "0.19.11"
app = marimo.App(width="medium")

with app.setup:
    import matplotlib.pyplot as plt
    import pymde
    import torch
    from wigglystuff import ChartSelect

    import marimo as mo

    pymde.seed(0)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    # Counties

    This notebook accompanies chapter 10 of the monograph [Minimum-Distortion Embedding](https://web.stanford.edu/~boyd/papers/min_dist_emb.html).

    In this eample notebook, we'll use PyMDE to embed and visualize 3,220 US counties, described by their demographic data (collected between 2013-2017 by an ACS longitudinal survey).

    We'll then color each county by the fraction of voters who voted for a democratic candidate in the 2016 presidential election. Interestingly, the embedding vaguely resembles a map of the US, though no geographic data was used to compute the embedding.
    """)
    return


@app.cell
def _():
    dataset = pymde.datasets.counties()
    return (dataset,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    Below is the raw dataframe, which was preprocessed using the code in scripts/preprocess_counties_data.py.

    The preprocessed data is stored in dataset.data
    """)
    return


@app.cell
def _(dataset):
    dataset.county_dataframe
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    We now make a neighbor-preserving embedding, to explore the local relationships in the data.
    """)
    return


@app.cell
def _(dataset):
    with mo.persistent_cache("counties"):
        mde = pymde.preserve_neighbors(data=dataset.data, verbose=True)
        embedding = mde.embed()
    return (embedding,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    Finally we visualize the embedding, rotating it so that it vaguely resembles a map of the US.

    Note that counties that voted Republican tend to cluster together, as do counties that voted Democratic.
    """)
    return


@app.cell
def _(dataset, embedding):
    # Rotate the embedding by some amount of degrees
    rotated_embedding = pymde.rotate(embedding, 150)
    ax = pymde.plot(
        rotated_embedding,
        color_by=dataset.attributes["democratic_fraction"],
        color_map="RdBu",
        marker_size=10,
    )
    plt.tight_layout()
    chart = mo.ui.anywidget(ChartSelect(ax.figure))
    chart
    return chart, rotated_embedding


@app.cell
def _(chart, rotated_embedding):
    selected_indices = chart.get_indices(rotated_embedding[:, 0], rotated_embedding[:, 1])
    return (selected_indices,)


@app.cell(hide_code=True)
def _(dataset, selected_indices):
    mo.stop(not selected_indices.size)

    selected_counties = dataset.county_dataframe.iloc[selected_indices]
    table = mo.ui.table(selected_counties)

    mo.output.replace(
        mo.md(
            f"""
        **You've selected {len(selected_indices)} counties.**

        {table}
        """
        )
    )
    return


if __name__ == "__main__":
    app.run()



# /// script
# requires-python = ">=3.10"
# dependencies = [
#     "marimo",
#     "pymde==0.2.3",
#     "torch==2.10.0",
# ]
# ///

import marimo

__generated_with = "0.19.11"
app = marimo.App()

with app.setup:
    import math
    import marimo as mo
    import pymde
    import torch


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    # Anchor Constraints

    In this notebook, we show how to use an [anchor constraint](https://pymde.org/mde/index.html#anchored), which
    lets us pin some of the embedding vectors to specific values. This can be useful if you know a priori what some of the embedding vectors should be.

    Below, we'll use an anchor constraint when laying out a binary tree: we'll pin the leaves to lie on a circle.
    """)
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    This next cell constraints a binary tree with a specified depth. You can change the depth of the tree by changing the value of the `depth` variable.
    """)
    return


@app.cell
def _():
    depth = 9
    n_items = 2 ** (depth + 1) - 1

    edges = []
    stack = [0]
    while stack:
        root = stack.pop()
        first_child = root * 2 + 1
        second_child = root * 2 + 2
        if first_child < n_items:
            edges.append((root, first_child))
            stack.append(first_child)
        if second_child < n_items:
            edges.append((root, second_child))
            stack.append(second_child)

    tree = pymde.Graph.from_edges(torch.tensor(edges))
    return depth, edges, n_items, tree


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    To layout the graph, we're going to preserve the shortest path distances between all pairs of nodes. The next cell computes these distances.
    """)
    return


@app.cell
def _(tree):
    shortest_paths_graph = pymde.preprocess.graph.shortest_paths(tree)
    return (shortest_paths_graph,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    To use an anchor constraint, we need the indices of the items we want to pin (remember that we will pin the leaves). We get the indices of the leaves using a simple formula.
    """)
    return


@app.cell
def _(depth):
    # these are the indices of the nodes that we will pin in place
    leaves = torch.arange(2**depth) + 2**depth - 1
    return (leaves,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    We want to pin the leaves to be uniformly spaced on a circle. The below cell constructs a Tensor in which the `i`th row holds the location to which the `i`th leaf should be pinned.

    If you like, you can try changing the `radius` to see what effect it has on the embedding.
    """)
    return


@app.cell
def _(leaves):
    radius = 20

    # pin the root to be at (0, 0), and the leaves to be spaced uniformly on a circle
    angles = torch.linspace(0, 2 * math.pi, leaves.numel() + 1)[1:]
    positions = radius * torch.stack([torch.sin(angles), torch.cos(angles)], dim=1)
    positions
    return (positions,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    Now that we have the leaves (the anchors) and the desired positions (their values), we can create the anchor constraint.
    """)
    return


@app.cell
def _(leaves, positions):
    anchor_constraint = pymde.Anchored(anchors=leaves, values=positions)
    return (anchor_constraint,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    We can now create the embedding. Feel free to play around with the distortion function (try changing it to `pymde.losses.Quadratic`, or `pymde.losses.Cubic`) to see what effect the choice of distortion function has on the emedding.
    """)
    return


@app.cell
def _():
    distortion_function = mo.ui.dropdown(
        {
            "WeightedQuadratic": pymde.losses.WeightedQuadratic,
            "Quadratic": pymde.losses.Quadratic,
            "Cubic": pymde.losses.Cubic,
        }, value="WeightedQuadratic"
    )
    distortion_function
    return (distortion_function,)


@app.cell
def _(
    anchor_constraint,
    distortion_function,
    edges,
    n_items,
    shortest_paths_graph,
):
    mde = pymde.MDE(
        n_items,
        embedding_dim=2,
        edges=shortest_paths_graph.edges,
        distortion_function=distortion_function.value(
            shortest_paths_graph.distances
        ),
        constraint=anchor_constraint,
    )
    mde.embed(snapshot_every=1, verbose=True, max_iter=15)
    mde.plot(edges=edges)
    return (mde,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    For fun, we can make a GIF of how the embedding was created, using the `play` method.
    """)
    return


@app.cell
def _(edges, mde):
    mde.play(edges=edges, savepath="/tmp/anchor.gif")
    mo.image("/tmp/anchor.gif")
    return


if __name__ == "__main__":
    app.run()



# /// script
# requires-python = ">=3.10"
# dependencies = [
#     "marimo",
#     "pymde",
#     "matplotlib",
#     "torch",
# ]
# ///

import marimo

__generated_with = "0.19.10"
app = marimo.App()

with app.setup:
    import matplotlib.pyplot as plt
    import marimo as mo
    import pymde
    import torch


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    # Fashion-MNIST

    This notebook shows how to use the function `pymde.preserve_neighbors` to produce embeddings that highlight the local structure of your data, using [Fashion-MNIST](https://github.com/zalandoresearch/fashion-mnist) as a case study. In these embeddings similar fashion articles from a class of 10 are near each other, and dissimilar articles are not near each other. Recall the class labels 0 till 9 correspond to:

    | Label | Description |
    | --- | --- |
    | 0 | T-shirt/top |
    | 1 | Trouser |
    | 2 | Pullover |
    | 3 | Dress |
    | 4 | Coat |
    | 5 | Sandal |
    | 6 | Shirt |
    | 7 | Sneaker |
    | 8 | Bag |
    | 9 | Ankle boot |

    It also shows how to debug embeddings and search for outliers in your original data, using PyMDE.
    """)
    return


@app.cell
def _():
    f_mnist = pymde.datasets.FashionMNIST()
    return (f_mnist,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## Embeddings, in one line
    """)
    return


@app.cell
def _(f_mnist):
    _embedding = pymde.preserve_neighbors(f_mnist.data, verbose=True).embed(verbose=True)
    pymde.plot(_embedding, color_by=f_mnist.attributes['class'])
    return


@app.cell
def _(f_mnist):
    _embedding = pymde.preserve_neighbors(f_mnist.data, constraint=pymde.Standardized()).embed(verbose=True)
    pymde.plot(_embedding, color_by=f_mnist.attributes['class'])
    return


@app.cell
def _(f_mnist):
    _embedding = pymde.preserve_neighbors(f_mnist.data, attractive_penalty=pymde.penalties.Quadratic, repulsive_penalty=None).embed(verbose=True)
    pymde.plot(_embedding, color_by=f_mnist.attributes['class'])
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## Embeddings, from scratch
    """)
    return


@app.cell
def _(f_mnist):
    knn_graph = pymde.preprocess.k_nearest_neighbors(f_mnist.data, k=15, verbose=True)
    return (knn_graph,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ### Quadratic MDE problems
    """)
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    The `embed` method computes and returns an embedding.

    The embedding is also accessible via the `X` attribute of the `MDE` instance.
    """)
    return


@app.cell
def _(f_mnist, knn_graph):
    quadratic_mde = pymde.MDE(
        n_items=f_mnist.data.shape[0],
        embedding_dim=2,
        edges=knn_graph.edges,
        distortion_function=pymde.penalties.Quadratic(knn_graph.weights),
        constraint=pymde.Standardized())

    quadratic_mde.embed(verbose=True)
    return (quadratic_mde,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    After embedding, we can call the `plot` method to visualize the embedding (when `embedding_dim` $\leq 3$).

    The `color_by` keyword argument takes a length-$n$ list of attribute values associated with the items; the values are used to color the points, with each unique value getting its own color.

    Here, the attribute is the fashion article depicted by the image.
    """)
    return


@app.cell
def _(f_mnist, quadratic_mde):
    quadratic_mde.plot(color_by=f_mnist.attributes['class'])
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    Next, we fit an embedding with 3 dimensions. We can visualize 3D embeddings with `PyMDE`.
    """)
    return


@app.cell
def _(f_mnist, knn_graph):
    quadratic_mde_3d = pymde.MDE(
        n_items=f_mnist.data.shape[0],
        embedding_dim=3,
        edges=knn_graph.edges,
        distortion_function=pymde.penalties.Quadratic(knn_graph.weights),
        constraint=pymde.Standardized())

    quadratic_mde_3d.embed(verbose=True)
    quadratic_mde_3d.plot(color_by=f_mnist.attributes['class'])
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## Debugging the embedding
    """)
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ### The distribution of distortions

    We can visualize the cumulative distribution of distortions. We see that a small number of pairs are responsible for most of the embedding's distortion.
    """)
    return


@app.cell
def _(quadratic_mde):
    quadratic_mde.distortions_cdf()
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ### High distortion pairs

    It can be instructive to inspect the pairs with highest distortion. A pair, here one column of the grid, is a set of two images that our MDE problem was told are similar. In this case, some of the high-distortion pairs are possiblity misclassified articles, and don't appear very similar to each other.
    """)
    return


@app.cell
def _(f_mnist, quadratic_mde):
    pairs, distortions = quadratic_mde.high_distortion_pairs()
    outliers = pairs[:10]

    def plot_pairs(pairs):
        fig, axs = plt.subplots(2, pairs.shape[0], figsize=(15.0, 3.))
        for pair_index in range(pairs.shape[0]):
            i = pairs[pair_index][0]
            j = pairs[pair_index][1]
            im_i = f_mnist.data[i].reshape(28, 28)
            im_j = f_mnist.data[j].reshape(28, 28)
            axs[0][pair_index].imshow(im_i, cmap=plt.cm.get_cmap('gray').reversed())
            axs[0][pair_index].set_xticks([])
            axs[0][pair_index].set_yticks([])
            axs[1][pair_index].imshow(im_j, cmap=plt.cm.get_cmap('gray').reversed())
            axs[1][pair_index].set_xticks([])
            axs[1][pair_index].set_yticks([])
        plt.tight_layout()

    plot_pairs(outliers)
    return pairs, plot_pairs


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## Low distortion pairs

    In comparison, the pairs with the lowest distortion look like very reasonable pairs of images.
    """)
    return


@app.cell
def _(pairs, plot_pairs):
    low_distortion_pairs = pairs[-10:]
    plot_pairs(low_distortion_pairs)
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## Including dissimilar pairs

    The quadratic MDE problem used a standardization constraint to enforce the embedding to spread out, and only used pairs of similar items.

    Instead of only relying on the constraint, we can also design the distortion functions so that they discourage dissimilar items from being close. We do this by including some pairs of dissimilar items, in addition to the pairs of similar items.
    """)
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ### Form $\mathcal{E} = \mathcal{E}_\text{sim} \cup \mathcal{E}_\text{dis}$
    """)
    return


@app.cell
def _(f_mnist, knn_graph):
    similar_edges = knn_graph.edges
    dissimilar_edges = pymde.preprocess.dissimilar_edges(n_items=f_mnist.data.shape[0], num_edges=similar_edges.shape[0], similar_edges=similar_edges)
    edges = torch.cat([similar_edges, dissimilar_edges])
    weights = torch.cat([knn_graph.weights, -1.0 * torch.ones(dissimilar_edges.shape[0])])
    plt.hist(weights.numpy())
    plt.gca()
    return edges, weights


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ### Define the distortion function $f$
    """)
    return


@app.cell
def _(weights):
    f = pymde.penalties.PushAndPull(
        weights=weights,
        attractive_penalty=pymde.penalties.Log1p,
        repulsive_penalty=pymde.penalties.Log,
    )
    return (f,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ### A standardized embedding ...
    """)
    return


@app.cell
def _(edges, f, f_mnist, quadratic_mde):
    std_mde = pymde.MDE(
        n_items=f_mnist.data.shape[0],
        embedding_dim=2,
        edges=edges,
        distortion_function=f,
        constraint=pymde.Standardized(),
    )
    std_mde.embed(X=quadratic_mde.X, max_iter=400, verbose=True)
    std_mde.plot(color_by=f_mnist.attributes['class'])
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## ... and an unconstrained embedding
    """)
    return


@app.cell
def _(edges, f, f_mnist, quadratic_mde):
    unconstrained_mde = pymde.MDE(
        n_items=f_mnist.data.shape[0],
        embedding_dim=2,
        edges=edges,
        distortion_function=f,
    )
    unconstrained_mde.embed(X=quadratic_mde.X, verbose=True)
    unconstrained_mde.plot(color_by=f_mnist.attributes['class'])
    return


if __name__ == "__main__":
    app.run()



# /// script
# requires-python = ">=3.10"
# dependencies = [
#     "marimo",
#     "pandas<3.0.0",
#     "pymde",
# ]
# ///

import marimo

__generated_with = "0.19.10"
app = marimo.App()

with app.setup:
    import marimo as mo
    import pymde


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    # Population genetics

    This notebook accompanies chapter 11 of the monograph [Minimum-Distortion Embedding](https://web.stanford.edu/~boyd/papers/min_dist_emb.html).

    In this eample notebook, we'll embed and visualize genomic data for 1,387 people thought to be of European ancestry, along with 154 synthetic (corrupted) points. We'll see that the embedding vaguely resembles the map of the Europe --- and that the synthetic points are isolated in their own cluster --- suggesting that the embedding is robust to noise, and that the genetic data encodes geographical data.
    """)
    return


@app.cell
def _():
    dataset = pymde.datasets.population_genetics()
    return (dataset,)


@app.cell
def _(dataset):
    mde = pymde.preserve_neighbors(
        data=dataset.corrupted_data,
        attractive_penalty=pymde.penalties.Huber,
        repulsive_penalty=pymde.penalties.Log,
        verbose=True
    )
    embedding = mde.embed()
    return (embedding,)


@app.cell
def _(dataset, embedding):
    pymde.plot(embedding, colors=dataset.attributes['corrupted_colors'], marker_size=10)
    return


if __name__ == "__main__":
    app.run()



# /// script
# requires-python = ">=3.14"
# dependencies = [
#     "marimo",
#     "matplotlib==3.10.8",
#     "numpy==2.3.0",
#     "plotly",
#     "pymde==0.2.3",
#     "torch==2.10.0",
# ]
# ///

import marimo

__generated_with = "0.19.11"
app = marimo.App()

with app.setup:
    import matplotlib.pyplot as plt
    import marimo as mo
    import numpy as np
    import plotly.graph_objects as go
    import pymde
    import torch


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    # Word embedding

    This notebook shows how to use PyMDE to create word embeddings in a few lines of code.

    We will embed the 5000 most popular academic interests listed in Google Scholar.
    """)
    return


@app.cell
def _():
    np.random.seed(0)
    torch.manual_seed(0)
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## Cooccurrence matrix

    The data is a **cooccurrence matrix**, which counts the number of times each pair of interests cooccurred in a researchers Google Scholar profile page.

    The ith row (and column) of the matrix corresponds to the ith interest (i.e., `interests[i]`)
    """)
    return


@app.cell
def _():
    dataset = pymde.datasets.google_scholar_interests()

    cooccurrence_graph = dataset.data
    interests = dataset.attributes["interests"]
    return cooccurrence_graph, interests


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    We can visualize the cooccurrence matrix. No patterns are immediately discernible.
    """)
    return


@app.cell
def _(cooccurrence_graph):
    plt.figure(figsize=(12, 12))
    plt.spy(cooccurrence_graph.A, ms=0.01, color="k")
    plt.gca()
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    As a first cut, we will embed the words into $\mathbf{R}^1$, using PCA.
    """)
    return


@app.cell
def _(cooccurrence_graph):
    pca_embedding = pymde.pca(cooccurrence_graph.A.todense(), 1)
    pca_embedding
    return (pca_embedding,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    We can visualize the PCA embedding by sorting the cooccurrence matrix according to the indices that sort the PCA embedding. Notice that this reveals some interesting structure.
    """)
    return


@app.cell
def _(cooccurrence_graph, pca_embedding):
    sort_indices = pca_embedding.flatten().sort().indices.numpy()
    A_sorted = cooccurrence_graph.A[sort_indices][:, sort_indices]

    plt.figure(figsize=(12, 12))
    plt.spy(A_sorted, ms=1e-2, color="k")
    plt.gca()
    return (sort_indices,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    Here are the first 20 entries in the sorted cooccurrence matrix. We can see that theey are mostly about artificial intelligence and biology, meaning that our simple embedding has uncovered that these topics are related.
    """)
    return


@app.cell
def _(interests, sort_indices):
    list(np.array(interests)[sort_indices])[:20]
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    ## Embedding

    Now we will construct a more sophisticated embedding. We start by converting the counts into dissimilarity scores,
    then compute a shortest path metric with respect to these scores.
    """)
    return


@app.cell
def _(cooccurrence_graph):
    dissimilarities = 1.0 / torch.log(cooccurrence_graph.weights)
    dissimilarity_graph = pymde.Graph.from_edges(
        cooccurrence_graph.edges, dissimilarities
    )
    return (dissimilarity_graph,)


@app.cell
def _(dissimilarity_graph):
    shortest_path_graph = pymde.preprocess.graph.shortest_paths(
        dissimilarity_graph, verbose=True, n_workers=6
    )
    return (shortest_path_graph,)


@app.cell
def _(interests):
    def interactive(X):
        if isinstance(X, torch.Tensor):
            X = X.cpu().numpy()
        fig = go.Figure(
            data=go.Scatter(
                x=X[:, 0],
                y=X[:, 1],
                marker_size=2.0,
                mode="markers",
                text=interests,
            )
        )
        fig.show()

    return (interactive,)


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    Finally we compute an embedding that preserves neighbors under the shortest path metric.
    """)
    return


@app.cell
def _(shortest_path_graph):
    mde = pymde.preserve_neighbors(shortest_path_graph, verbose=True)
    return (mde,)


@app.cell
def _(mde):
    X = mde.embed(verbose=True)
    return (X,)


@app.cell
def _(mde):
    mde.plot(colors=["black"])
    plt.gca()
    return


@app.cell(hide_code=True)
def _():
    mo.md(r"""
    Try mousing over this embedding, to reveal the interest labels for each node. You'll find that the embedding has done a good job of clustering similar interests together.
    """)
    return


@app.cell
def _(X, interactive):
    interactive(X)
    return


if __name__ == "__main__":
    app.run()
